"""
This Python module is intended for use with the LLDB debugger.

To import using this module in an LLDB session, execute the following command
in LLDB::

    command script import <path-to-lldb_bson.py>

This may be useful to add as a "startup/init" command for an LLDB executor such
as VSCode.

From this point, all ``bson_t`` variables will show their full contents
recursively.

- Bonus: If using CodeLLDB, it is possible to add watch expressions to elements
  within a (nested) BSON document.

  To add a Python watch expression, prefix the watch string with ``/py``, and
  the remainder of the string will be evaluated in the LLDB Python interpreter
  rather than the LLDB evaluator. For example:

    /py 'hello, ' + 'world'

  will evaluate to a string "Hello, world".

  To reference a bson_t object, use the ``$variable_name`` syntax. Then, use the
  ``@`` operator with the special global ``bson`` object to walk through the
  document as if it were a regular Python object::

    /py $my_bson_t @ bson.topElement.path["string key"].array[3].name

  Only the leading ``bson`` is required in the path, since it provides the
  "magic" necessary to perform document traversal. For example, with the
  given document ``userdata``::

    {
        users: [
            {username: "joe", age: 41},
            {username: "alex", age: 33},
            {username: "jane', age: 29}
        ]
    }

  the "age" of "alex" can be watched with the following watch expression:

    /py $userdata @bson.users[1].age

"""

from __future__ import annotations

import enum
import functools
import hashlib
import json
import shlex
import struct
import traceback
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    Generator,
    Generic,
    Iterable,
    NamedTuple,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

import lldb  # pyright: ignore
from lldb import (  # pyright: ignore
    SBAddress,
    SBDebugger,
    SBError,
    SBFrame,
    SBProcess,
    SBSyntheticValueProvider,
    SBType,
    SBValue,
)

if TYPE_CHECKING:
    from typing_extensions import override
else:

    def override(f: T) -> T:
        return f


def print_errors(fn: FuncT) -> FuncT:
    """Exceptions from the decorated function will be printed, then re-raised"""

    @functools.wraps(fn)
    def _wrap(*args: Any, **kwargs: Any) -> Any:
        try:
            return fn(*args, **kwargs)
        except:
            e = traceback.format_exc()
            print(e)
            raise

    return cast("FuncT", _wrap)


@print_errors
def __lldb_init_module(debugger: SBDebugger, internal_dict: InternalDict):
    # Inject the global magic document traverser:
    internal_dict["bson"] = _BSONWalker()
    # Register types:
    for cls in _SyntheticMeta.synthetics:
        # The (regex of) the type that is handled by this class:
        ty = shlex.quote(cls.__typename__)
        if cls.__enable_synthetic__:
            # Register a synthetic child provider, generating the trees that can be expanded
            cmd = f"type synthetic add -l lldb_bson.{cls.__name__} -x '^{ty}$'"
            debugger.HandleCommand(cmd)
        if cls.__summary_str__ is not None:
            # Generate a simple summary string for this object
            quoted = cls.__summary_str__.replace("'", "\\'")
            cmd = f"type summary add --summary-string '{quoted}' -x '^{ty}$'"
            debugger.HandleCommand(cmd)
        if hasattr(cls, "__summary__"):
            # More complex: Call a Python function that will create the summary
            cmd = f"type summary add -F lldb_bson.{cls.__name__}.__summary__ -x '^{ty}$'"
            debugger.HandleCommand(cmd)

    # Render __bson_byte__ as "bytes with ASCII." __bson_byte__ is a
    # debug-only type generated on-the-fly in LLDB
    debugger.HandleCommand("type format add -f Y __bson_byte__")
    # Arrays of bytes as a sequence of hex values:
    debugger.HandleCommand(r"type summary add -s '${var[]%x}' -x '__bson_byte__\[[0-9]+\]'")

    print("lldb_bson is ready")


_ = __lldb_init_module  # Silence "unused function" warnings


FuncT = TypeVar("FuncT", bound=Callable[..., Any])
"Type of functions"
T = TypeVar("T")
"Unbounded invariant type parameter"
InternalDict = Dict[str, Any]
"Type of internal dictionaries, provided by LLDB"


ValueFactory = Callable[[], SBValue]
ChildItem = Union[
    Tuple[str, "str | int"], ValueFactory, Tuple[str, "str | int", "lldb.ValueFormatType|None", "SBType|None"]
]


class _SyntheticMeta(type):
    """
    Metaclass that handles subclassing of SyntheticDisplayBase. Does basic checks
    and collects definitions into a list of display types
    """

    synthetics: list[Type[SyntheticDisplayBase[Any]]] = []
    "The display type classes that have been defined"

    @override
    def __new__(
        cls: Type[_SyntheticMeta], name: str, bases: tuple[type, ...], namespace: dict[str, Any]
    ) -> Type[SyntheticDisplayBase[Any]]:
        new_class: Type[SyntheticDisplayBase[Any]] = type.__new__(cast(type, cls), name, bases, namespace)
        if namespace.get("__abstract__"):
            return new_class
        # Check for the required __typename__ and __parse__
        if not hasattr(new_class, "__typename__"):
            raise TypeError(f'Type "{new_class}" is missing a "__typename__" attribute')
        if not hasattr(new_class, "__parse__"):
            raise TypeError(f'Type "{new_class}" has no "__parse__" method')
        # Remember this new class:
        cls.synthetics.append(new_class)
        return new_class


class SyntheticDisplayBase(Generic[T], SBSyntheticValueProvider, metaclass=_SyntheticMeta):
    __abstract__: ClassVar[bool] = True
    "If true, disables metaclass checks"

    __summary_str__: ClassVar[str | None] = None
    "Set to an LLDB '--summary-string' formatting string for rendering the inline value summary"
    __enable_synthetic__: ClassVar[bool] = True
    "If False, do not generate synthetic children (used for primitive values)"

    if TYPE_CHECKING:
        __typename__: ClassVar[str]
        """
        The (regular expression) name of the type that is handled by this class.
        This is a required class attribute.
        """

        @classmethod
        def __summary__(cls, value: SBValue, idict: InternalDict) -> str:
            """
            If provided, supplies the Python function that will dynamically
            generate a summary string from the SBValue.
            """
            ...

        @classmethod
        def __parse__(cls, value: SBValue) -> T:
            """
            Required method: Parse the data that lives at the given SBValue into
            a representation that will be useful for later display.
            """
            ...

    @classmethod
    @print_errors
    def __get_sbtype__(cls, frame: SBFrame, addr: int) -> SBType:
        """
        Obtain the SBType for this class. Can be overriden in subclasses, and
        the type may consider the value that lives at the address.
        """
        return generate_or_get_type(f"struct {cls.__typename__} {{}}", frame)

    @print_errors
    def __init__(self, val: SBValue, idict: InternalDict | None = None) -> None:
        self.__sbvalue = val
        "The SBValue given for this object"
        self.__children: list[ChildItem] = []
        "The synthetic children associated with the value"
        self.__value: T | None = None
        "The decoded value, or ``None`` if it has not yet been decoded"

    @property
    def sbvalue(self) -> SBValue:
        """Obtain the SBValue associated with this display object"""
        return self.__sbvalue

    @property
    def value(self) -> T:
        """Obtain the decoded BSON value"""
        if self.__value is None:
            self.__value = self.__parse__(self.sbvalue)
        return self.__value

    @property
    def address(self) -> int:
        """Obtain the memory address of the associated object"""
        return self.sbvalue.load_addr

    @print_errors
    @override
    def update(self) -> bool | None:
        """Update the parsed value and the child objects"""
        self.__value = None  # Clear the value so it will be re-read
        self.__children = list(self.get_children())

    def get_children(self) -> Iterable[ChildItem]:
        """
        Subclasses must provide this method such that it returns an iterable
        object of ``ChildItem`` types. These will be used to generate the
        synthetic children that are displayed by LLDB.
        """
        raise NotImplementedError

    @override
    def num_children(self) -> int:
        """Called by LLDB to know how many children exist"""
        n = len(self.__children)
        return n

    @override
    def has_children(self) -> bool:
        """Optimization opportunity for LLDB if it knows it doesn't need to ask"""
        return self.__enable_synthetic__ and self.num_children() != 0

    @print_errors
    @override
    def get_child_at_index(self, pos: int) -> SBValue:
        """
        Obtain the synthetic child of this value at index 'pos'.
        """
        # LLDB sometimes calls us with a child that we don't have?
        if pos >= len(self.__children):
            print(f"NOTE: lldb called get_child_at_index({pos}), but we only have {len(self.__children)} children")
            return SBValue()
        # Get the child:
        nth = self.__children[pos]
        # If not a tuple:
        if not isinstance(nth, tuple):
            # The type is a ValueFactory, which will return a new SBValue
            val = nth()
            assert val.error.success, f"{val.error=}, {nth=}, {pos=}"
            return val
        # Otherwise, they yielded a tuple:
        if len(nth) == 4:
            # A four-tuple of key, value, an LLDB value format, and an LLDB value type
            key, val, fmt, ty = nth
            sbval = self.create_value(key, val, format=fmt, cast_to=ty)
        else:
            # Just a key and a value
            key, val = nth
            sbval = self.create_value(key, val)
        assert sbval.error.success, (sbval.error, key, val)
        return sbval

    def create_value(
        self,
        name: str,
        value: str | int | float,
        cast_to: SBType | None = None,
        format: lldb.ValueFormatType | None = None,
    ) -> SBValue:
        """Cast a Python primitive into an LLDB SBValue."""
        # Encode the value in a JSON string, which will coincidentally be a valid
        # C expression that LLDB can parse:
        encoded = json.dumps(value)
        val = self.__sbvalue.CreateValueFromExpression(name, encoded)
        if cast_to:
            val = val.Cast(cast_to)
        if format is not None:
            val.format = format
        return val


class PrimitiveDisplay(Generic[T], SyntheticDisplayBase[T]):
    """
    Displays a primitive type. We can't use LLDB's basic types directly, since
    we must force a little-endian encoding.

    Set the __struct_format__ class variable to a Python struct format string.
    """

    __abstract__ = True
    __enable_synthetic__: ClassVar[bool] = False

    __struct_format__: ClassVar[str]
    "The struct format string that will be used to extract the value from memory"

    @classmethod
    @override
    @print_errors
    def __summary__(cls, value: SBValue, idict: InternalDict) -> str:
        return json.dumps(cls.__parse__(value))

    @classmethod
    @override
    def __parse__(cls, value: SBValue) -> T:
        unpack = struct.Struct(cls.__struct_format__)
        buf = memcache.get_cached(value.load_addr)[: unpack.size]
        val: T = unpack.unpack(buf)[0]
        return val


class DoubleDisplay(PrimitiveDisplay[float]):
    """Displays BSON doubles"""

    __typename__ = "__bson_double__"
    __struct_format__: ClassVar[str] = "<d"


class UTF8Display(SyntheticDisplayBase[bytes]):
    """Display type for BSON UTF-8 values"""

    __typename__ = "__bson_utf8__"
    __summary_str__ = "${var[1]}"  # Display the second child (which is the actual string)

    @classmethod
    @override
    def __parse__(cls, value: SBValue) -> bytes:
        buf = memcache.get_cached(value.load_addr)
        size = read_i32le(buf)
        return bytes(buf[4:][:size])

    @override
    def get_children(self) -> Iterable[ChildItem]:
        strlen = len(self.value)
        yield "size (bytes)", strlen
        # Create a char[] type to represent the string content:
        array_t = self.sbvalue.target.GetBasicType(lldb.eBasicTypeChar).GetArrayType(strlen)
        yield lambda: self.sbvalue.synthetic_child_from_address("[content]", self.address + 4, array_t)
        try:
            # Attempt a UTF-8 decode. We don't actually show this, we just want to
            # check if there are encoding errors, which we will display in the output
            self.value.decode("utf-8")
        except UnicodeDecodeError as e:
            yield "decode error", str(e)


class DocumentInfo(NamedTuple):
    """A decoded document"""

    elements: Sequence[DocumentElement | DocumentError]
    "Existing elements or errors found while parsing the data"


class DocumentElement(NamedTuple):
    """Represents an element within a document"""

    type: BSONType
    key: str
    value_offset: int
    "Offset from the beginning of the document data where the element's value appears"
    value_size: int
    "The size of the element's value (in bytes)"


class DocumentError(NamedTuple):
    """Represents an error while decoding a document"""

    message: str
    error_offset: int


class DocumentDisplay(SyntheticDisplayBase["DocumentInfo | DocumentError"]):
    """
    Main display of BSON document elements. This parses a document/array, and
    generates the child elements that can be further expanded and inspected.

    This type does not refer to the _MemCache, since this object is usually
    the top-level object and is the one responsible for filling the cache.
    """

    __typename__ = "__bson_document_[0-9]+__"
    __qualifier__: ClassVar[str] = "document"
    "The 'qualifier' of this type. Overriden by ArrayDisplay."

    @classmethod
    @override
    @print_errors
    def __summary__(cls, value: SBValue, idict: InternalDict) -> str:
        prefix = cls.__qualifier__
        doc = cls.__parse__(value)
        if isinstance(doc, DocumentError):
            return f"Error parsing {prefix} at byte {doc.error_offset}: {doc.message}"
        if len(doc.elements) == 0:
            return f"{prefix} (empty)"
        if len(doc.elements) == 1:
            return f"{prefix} (1 element)"
        return f"{prefix} ({len(doc.elements)} elements)"

    @classmethod
    @override
    def __get_sbtype__(cls, frame: SBFrame, addr: int) -> SBType:
        """Generate a unique type for the length of the document, allowing for byte-wise inspection"""
        # Read the size prefix:
        err = SBError()
        header = frame.thread.process.ReadMemory(addr, 4, err)
        assert err.success, f"{err=}, {frame=}, {addr=}"
        size = read_i32le(header)
        # Generate the type:
        typename = f"__bson_{cls.__qualifier__}_{size}__"
        doc_t = generate_or_get_type(
            f"""
            enum __bson_byte__ : unsigned char {{}};
            struct {typename} {{ __bson_byte__ bytes[{size}]; }}
            """,
            frame,
        )
        return doc_t

    @classmethod
    @override
    def __parse__(cls, value: SBValue) -> DocumentInfo | DocumentError:
        try:
            # Read from memory and refresh the segment in the memory cache. All
            # child element displays will re-use the same memory segment that
            # will be pulled here:
            buf = memcache.read(value)[1]
        except LookupError as e:
            return DocumentError(f"Failed to read memory: {e}", value.load_addr)
        return cls.parse_bytes(buf)

    @classmethod
    def parse_bytes(cls, buf: bytes) -> DocumentInfo | DocumentError:
        """Parse a document from the given data buffer"""
        elems = list(cls._parse_elems(buf))
        return DocumentInfo(elems)

    @classmethod
    def _parse_elems(cls, buf: bytes) -> Iterable[DocumentElement | DocumentError]:
        """Iteratively yield elements, or an error if parsing fails"""
        cur_offset = 4
        array_idx = 0
        while buf:
            elem = yield from cls._parse_one(buf[cur_offset:], cur_offset)
            if isinstance(elem, DocumentError):
                # An error ocurred, so we can't reliably continue parsing
                yield elem
                return
            if elem.type == BSONType.EOD:
                # This is the end.
                break
            # Yield this one, and then advance to the next element:
            yield elem
            elem_size = 1 + len(elem.key) + 1 + elem.value_size
            if cls.__qualifier__ == "array":
                # Validate that array keys are increasing integers:
                expect_key = str(array_idx)
                if elem.key != expect_key:
                    yield DocumentError(
                        f"Array element must have incrementing integer keys "
                        f'(Expected "{expect_key}", got "{elem.key}")',
                        cur_offset,
                    )
            array_idx += 1
            cur_offset += elem_size
        # Check that we actually consumed the whole buffer:
        remain = len(buf) - cur_offset
        if remain > 1:
            yield DocumentError(f"Extra {len(buf)} bytes in document data", cur_offset)

    @classmethod
    def _parse_one(
        cls, buf: bytes, elem_offset: int
    ) -> Generator[DocumentError, None, DocumentElement | DocumentError]:
        try:
            # Read the tag type
            type_tag = BSONType(buf[0])
        except ValueError:
            # The tag byte is not a valid tag value
            return DocumentError(f"Invalid element type tag 0x{buf[0]:x}", elem_offset)
        except IndexError:
            # 'buf' was empty
            return DocumentError(f"Unexpected end-of-data", elem_offset)
        # Stop if this is the end:
        if type_tag == BSONType.EOD:
            return DocumentElement(type_tag, "", 0, 0)
        # Find the null terminator on the key:
        try:
            key_nulpos = buf.index(0, 1)
        except ValueError:
            return DocumentError(f"Unexpected end-of-data while parsing the element key", elem_offset)
        key_bytes = buf[1:key_nulpos]
        try:
            key = key_bytes.decode("utf-8")
        except UnicodeDecodeError as e:
            yield DocumentError(f"Element key {key_bytes} is not valid UTF-8 ({e})", elem_offset)
            key = key_bytes.decode("utf-8", errors="replace")
        # The offset of the value within the element:
        inner_offset = key_nulpos + 1
        # The buffer that starts at the value:
        value_bytes = buf[inner_offset:]
        # Get the fixed size of the element:
        fixed_size = {
            BSONType.Double: 8,
            BSONType.Bool: 1,
            BSONType.Undefined: 0,
            BSONType.Null: 0,
            BSONType.MinKey: 0,
            BSONType.MaxKey: 0,
            BSONType.Datetime: 8,
            BSONType.ObjectID: 12,
            BSONType.Int32: 4,
            BSONType.Timestamp: 8,
            BSONType.Int64: 8,
            BSONType.Decimal128: 16,
        }
        value_size = fixed_size.get(BSONType(type_tag))
        if value_size is not None:
            pass  # This element has a fixed size
        elif type_tag in (BSONType.Code, BSONType.Symbol, BSONType.UTF8):
            # Size is 4 + a length prefix
            value_size = read_i32le(value_bytes) + 4
        elif type_tag in (BSONType.Array, BSONType.Document, BSONType.CodeWithScope):
            # Size is given by the length prefix
            value_size = read_i32le(value_bytes)
        elif type_tag == BSONType.DBPointer:
            # Size is a length prefix, plus four, plus 12 bytes for the OID
            value_size = read_i32le(value_bytes) + 4 + 12
        elif type_tag == BSONType.Regex:
            # Size is dynamic and given as two C strings:
            nul1 = value_bytes.index(0)
            nul2 = value_bytes.index(0, nul1 + 1)
            value_size = nul2 + 1
        elif type_tag == BSONType.Binary:
            # Size is a length prefix, plus four, plus one for the subtype
            value_size = read_i32le(value_bytes) + 4 + 1
        else:
            assert False, f"Unhandled value tag? {type_tag=} {buf=} {key=}"
        # The absolute offset of the element within the parent document:
        value_offset = elem_offset + inner_offset
        return DocumentElement(type_tag, key, value_offset, value_size)

    @override
    def get_children(self) -> Iterable[ChildItem]:
        doc = self.value
        if isinstance(doc, DocumentError):
            # The entire document failed to parse. Just generate one error:
            yield "[error]", f"Parsing error at byte {doc.error_offset}: {doc.message}"
            return
        for elem in doc.elements:
            if isinstance(elem, DocumentError):
                # There was an error at this location.
                yield "[error]", f"Data error at offset {elem.error_offset}: {elem.message}"
            else:
                # Create a ValueFactory for each element:
                yield functools.partial(self.create_child, self.sbvalue, elem)

    @classmethod
    def create_child(cls, parent: SBValue, elem: DocumentElement) -> SBValue:
        """Generate the child elements for LLDB to walk through"""
        if cls.__qualifier__ == "array":
            # Don't quote the integer keys
            name = f"[{elem.key}]"
        else:
            name = f"['{elem.key}']"
        value_addr = parent.load_addr + elem.value_offset
        frame = parent.frame
        # Create a new SBType to represent the element value. For each type tag, we
        # want a function that maps an SBFrame and an address to a new SBType:
        by_type: dict[BSONType, Callable[[SBFrame, int], SBType]] = {
            BSONType.Double: DoubleDisplay.__get_sbtype__,
            BSONType.UTF8: UTF8Display.__get_sbtype__,
            BSONType.Document: DocumentDisplay.__get_sbtype__,
            BSONType.Binary: BinaryDisplay.__get_sbtype__,
            BSONType.Array: ArrayDisplay.__get_sbtype__,
            # For bool, we can just use LLDB's basic type:
            BSONType.Bool: lambda _fr, _addr: parent.target.GetBasicType(lldb.eBasicTypeBool),
            BSONType.Code: CodeDisplay.__get_sbtype__,
            BSONType.CodeWithScope: CodeWithScopeDisplay.__get_sbtype__,
            BSONType.Int32: Int32Display.__get_sbtype__,
            BSONType.Int64: Int64Display.__get_sbtype__,
            BSONType.ObjectID: ObjectIDDisplay.__get_sbtype__,
            BSONType.DBPointer: DBPointerDisplay.__get_sbtype__,
            BSONType.Regex: RegexDisplay.__get_sbtype__,
            BSONType.Symbol: SymbolDisplay.__get_sbtype__,
            BSONType.Datetime: DatetimeDisplay.__get_sbtype__,
            BSONType.Timestamp: TimestampDisplay.__get_sbtype__,
            BSONType.Decimal128: Decimal128Display.__get_sbtype__,
            BSONType.Null: NullDisplay.__get_sbtype__,
            BSONType.Undefined: UndefinedDisplay.__get_sbtype__,
            BSONType.MaxKey: MaxKeyDisplay.__get_sbtype__,
            BSONType.MinKey: MinKeyDisplay.__get_sbtype__,
        }
        get_type = by_type.get(elem.type)
        assert get_type is not None, f"Unhandled type tag? {elem=}"
        # Create the SBType:
        type = get_type(frame, value_addr)
        # Create a synthetic child of that type at the address of the element's value:
        val = parent.synthetic_child_from_address(name, value_addr, type)
        assert val.error.success, f"{elem=}, {val.error=}"
        return val


class ArrayDisplay(DocumentDisplay):
    """Display for arrays. Most logic is implemented in the DocumentDisplay base."""

    __typename__ = "__bson_array_[0-9]+__"
    __qualifier__: ClassVar[str] = "array"


class BinaryInfo(NamedTuple):
    subtype: int
    data: bytes


class BinaryDisplay(SyntheticDisplayBase[BinaryInfo]):
    """Display for a BSON binary value"""

    __typename__ = "__bson_binary__"

    @classmethod
    @override
    def __parse__(cls, value: SBValue) -> BinaryInfo:
        buf = memcache.get_cached(value.load_addr)
        # Size prefix:
        data_size = read_i32le(buf)
        # Type tag:
        type = buf[4]
        # The remainder of the data:
        data = buf[5:][:data_size]
        return BinaryInfo(type, data)

    @override
    def get_children(self) -> Iterable[ChildItem]:
        yield "size", len(self.value.data)
        byte_t = generate_or_get_type("enum __bson_byte__ : char {}", self.sbvalue.frame)
        yield "subtype", self.value.subtype, lldb.eFormatHex, byte_t
        array_t = byte_t.GetArrayType(len(self.value.data))
        yield lambda: self.sbvalue.synthetic_child_from_address("data", self.address + 5, array_t)


class UndefinedDisplay(SyntheticDisplayBase[None]):
    """
    Display type for 'undefined' values. Also derived from for other unit types.
    """

    __typename__ = "__bson_undefined__"
    __summary_str__ = "undefined"
    __enable_synthetic__: ClassVar[bool] = False

    @classmethod
    @override
    def __parse__(cls, value: SBValue) -> None:
        return None


class ObjectIDDisplay(SyntheticDisplayBase[bytes]):
    """Display type for ObjectIDs"""

    __typename__ = "__bson_objectid__"

    @classmethod
    @override
    def __summary__(cls, value: SBValue, idict: InternalDict) -> str:
        val = cls.__parse__(value)
        return f'ObjectID("{val.hex()}")'

    @classmethod
    @override
    def __get_sbtype__(cls, frame: SBFrame, addr: int) -> SBType:
        return generate_or_get_type(
            r"""
            enum __bson_byte__ : char {};
            struct __bson_objectid__ { __bson_byte__ bytes[12]; }
            """,
            frame,
        )

    @classmethod
    @override
    def __parse__(cls, value: SBValue) -> bytes:
        buf = memcache.get_cached(value.load_addr)
        return buf[:12]

    @override
    def get_children(self) -> Iterable[ChildItem]:
        yield "spelling", self.value.hex()


class DatetimeDisplay(SyntheticDisplayBase[int]):
    """Display for BSON Datetime objects"""

    __typename__ = "__bson_datetime__"
    __summary_str__: ClassVar[str] = "datetime: ${var[0]}"

    @classmethod
    @override
    def __summary__(cls, value: SBValue, idict: InternalDict) -> str:
        dt = datetime.fromtimestamp(cls.__parse__(value) / 1000)
        s = f"{dt:%a %b %m %Y %H:%M:%S +%fμs}"
        return f'Date("{s}")'

    @classmethod
    @override
    def __parse__(cls, val: SBValue) -> int:
        buf = memcache.get_cached(val.load_addr)
        buf = buf[:8]
        value: int = struct.unpack("<Q", buf)[0]
        return value

    @override
    def get_children(self) -> Iterable[ChildItem]:
        # We can create a rich display using Python's datetime parsing:
        dt = datetime.fromtimestamp(self.value / 1000)
        # Adjusted to the local time zone:
        adjusted = dt.astimezone()
        yield from {
            "[isoformat]": dt.isoformat(),
            "[date]": f"{dt:%B %d, %Y}",
            "[time]": dt.strftime("%H:%M:%S +%fμs"),
            "[local]": adjusted.strftime("%c"),
            "Year": dt.year,
            "Month": dt.month,
            "Day": dt.day,
            "Hour": dt.hour,
            "Minute": dt.minute,
            "Second": dt.second,
            "+μs": dt.microsecond,
        }.items()


class NullDisplay(UndefinedDisplay):
    """Display for the BSON 'null' type"""

    __typename__ = "__bson_null__"
    __summary_str__ = "null"


class RegexDisplay(SyntheticDisplayBase[Tuple[bytes, bytes]]):
    """Display type for BSON regular expressions"""

    __typename__ = "__bson_regex_[0-9]+_[0-9]+__"
    __enable_synthetic__: ClassVar[bool] = False

    @classmethod
    @override
    def __get_sbtype__(cls, frame: SBFrame, addr: int) -> SBType:
        regex, opts = cls.parse_at(addr)
        regex_len = len(regex) + 1
        opts_len = len(opts) + 1
        # Synthesize a struct with two char[] matching the strings:
        return generate_or_get_type(
            rf"""
            struct __bson_regex_{regex_len}_{opts_len}__ {{
                char regex[{regex_len}];
                char options[{opts_len}];
            }}
            """,
            frame,
        )

    @classmethod
    @override
    def __parse__(cls, value: SBValue) -> tuple[bytes, bytes]:
        return cls.parse_at(value.load_addr)

    @classmethod
    @override
    def __summary__(cls, value: SBValue, idict: InternalDict) -> str:
        # Create a JS-style regex literal:
        pair = cls.__parse__(value)
        regex, options = cls.decode_pair(pair)
        regex = regex.replace("/", "\\/").replace("\n", "\\n")
        return f"/{regex}/{options}"

    @classmethod
    def parse_at(cls, addr: int) -> tuple[bytes, bytes]:
        buf = memcache.get_cached(addr)
        # A regex is encoded with two C-strings. Find the nulls:
        nulpos_1 = buf.index(0)
        nulpos_2 = buf.index(0, nulpos_1 + 1)
        # Split them up:
        regex = buf[:nulpos_1]
        options = buf[nulpos_1 + 1 : nulpos_2]
        # Decode and return:
        return regex, options

    @classmethod
    def decode_pair(cls, value: tuple[bytes, bytes]) -> tuple[str, str]:
        regex, options = value
        regex = regex.decode("utf-8", errors="replace")
        options = options.decode("utf-8", errors="replace")
        return regex, options


class DBPointerDisplay(SyntheticDisplayBase[Tuple[bytes, int]]):
    """Display type for DBPointers"""

    __typename__ = "__bson_dbpointer__"
    __summary_str__: ClassVar[str | None] = "DBPointer(${var[0]}, ${var[1]})"

    @classmethod
    @override
    def __parse__(cls, value: SBValue) -> tuple[bytes, int]:
        buf = memcache.get_cached(value.load_addr)
        # Grab the string and the OID position
        strlen = read_i32le(buf)
        size = strlen + 4 + 12
        return buf[:size], strlen + 4

    @override
    def get_children(self) -> Iterable[ChildItem]:
        utf8_t = UTF8Display.__get_sbtype__(self.sbvalue.frame, self.address)
        oid_offset = self.value[1]
        oid_t = ObjectIDDisplay.__get_sbtype__(self.sbvalue.frame, self.address + oid_offset)
        yield lambda: self.sbvalue.synthetic_child_from_address("collection", self.sbvalue.load_addr, utf8_t)
        yield lambda: self.sbvalue.synthetic_child_from_address("object", self.sbvalue.load_addr + oid_offset, oid_t)


class CodeDisplay(UTF8Display):
    """Display type for BSON code"""

    __typename__ = "__bson_code__"
    __summary_str__ = "Code(${var[1]})"


class SymbolDisplay(UTF8Display):
    """Display type for BSON symbols"""

    __typename__ = "__bson_symbol__"

    @classmethod
    @override
    def __summary__(cls, value: SBValue, idict: InternalDict) -> str:
        spell = cls.__parse__(value)
        dec = spell.decode("utf-8", errors="replace").rstrip("\x00")
        return f"Symbol({dec})"


class CodeWithScopeDisplay(SyntheticDisplayBase[int]):
    """Display type for BSON 'Code w/ Scope'"""

    __typename__ = "__code_with_scope__"

    __summary_str__: ClassVar[str | None] = "Code(${var[0][1]}, ${var[1]})"

    @classmethod
    @override
    def __parse__(cls, value: SBValue) -> int:
        buf = memcache.get_cached(value.load_addr)
        # Find the end position of the string:
        strlen = read_i32le(buf[4:])
        str_end = 4 + strlen + 4
        return str_end

    @override
    def get_children(self) -> Iterable[ChildItem]:
        code_t = CodeDisplay.__get_sbtype__(self.sbvalue.frame, self.address)
        scope_doc_offset = self.value
        doc_t = DocumentDisplay.__get_sbtype__(self.sbvalue.frame, self.address + scope_doc_offset)
        yield lambda: checked(self.sbvalue.synthetic_child_from_address("code", self.address + 4, code_t))
        yield lambda: checked(
            self.sbvalue.synthetic_child_from_address("scope", self.address + scope_doc_offset, doc_t)
        )


class Int32Display(PrimitiveDisplay[int]):
    """Display for 32-bit BSON integers"""

    __typename__ = "__bson_int32__"
    __struct_format__: ClassVar[str] = "<i"

    @classmethod
    @override
    def __summary__(cls, value: SBValue, idict: InternalDict) -> str:
        return f"NumberInt({cls.__parse__(value)})"


class Int64Display(PrimitiveDisplay[int]):
    """Display for 64-bit BSON integers"""

    __typename__ = "__bson_int64__"
    __struct_format__: ClassVar[str] = "<q"

    @classmethod
    @override
    def __summary__(cls, value: SBValue, idict: InternalDict) -> str:
        return f"NumberLong({cls.__parse__(value)})"


class TimestampDisplay(SyntheticDisplayBase[Tuple[int, int]]):
    """Display type for BSON timestamps"""

    __typename__ = "__bson_timestamp__"
    __summary_str__ = "Timestamp(${var[0]}, ${var[1]})"

    @classmethod
    @override
    def __parse__(cls, value: SBValue) -> tuple[int, int]:
        buf = memcache.get_cached(value.load_addr)[:8]
        # Just two 32bit integers:
        timestamp, increment = struct.unpack("<ii", buf)
        return timestamp, increment

    @override
    def get_children(self) -> Iterable[ChildItem]:
        yield "timestamp", self.value[0]
        yield "increment", self.value[1]


class Decimal128Value(NamedTuple):
    """Represents a parsed Decimal128 value"""

    sign: int
    combination: int
    exponent: int
    significand: int
    spelling: str


class Decimal128Display(SyntheticDisplayBase[Decimal128Value]):
    """The display type for BSON's Decimal128 type"""

    __typename__ = "__bson_decimal128__"

    @classmethod
    @override
    def __get_sbtype__(cls, frame: SBFrame, addr: int) -> SBType:
        """Generate a type for byte-wise introspection"""
        return generate_or_get_type(
            r"""
            struct __bson_decimal128__ {
                unsigned char bytes[16];
            }
            """,
            frame,
        )

    @classmethod
    @override
    def __summary__(cls, value: SBValue, idict: InternalDict) -> str:
        val = cls.__parse__(value)
        return f'NumberDecimal("{val.spelling}")'

    @classmethod
    @override
    def __parse__(cls, value: SBValue) -> Decimal128Value:
        dat = bytes(value.data.uint8)
        # The value is little-endian encoded, so we manually read it in that encoding
        d1 = read_i32le(dat)
        d2 = read_i32le(dat[4:])
        d3 = read_i32le(dat[8:])
        d4 = read_i32le(dat[12:])
        # Combin the parts:
        low_word = (d2 << 32) | d1
        hi_word = (d4 << 32) | d3
        d128_tetra = (hi_word << 64) | low_word
        # Create an array of individual bits (high bits first):
        bits = tuple(((d128_tetra >> n) & 1) for n in range(127, -1, -1))
        # Recombine a sequence of bits into an int (high bits first)
        mergebits: Callable[[tuple[int, ...]], int] = lambda bs: functools.reduce(lambda a, b: (a << 1) | b, bs, 0)
        # Sign bit:
        sign = bits[0]
        # BID uses the first two combo bits to indicate that the exponent is shifted
        if bits[1] and bits[2]:
            shifted = True
        else:
            shifted = False

        spelling = None
        if not shifted:
            # Regular bit positions:
            exponent = bits[1:15]
            coeff = bits[15:]
        else:
            # Bit positions are shifted over
            exponent = bits[3:17]
            # The significand has an implicit '0b100' prepended as the highest bits:
            coeff = (1, 0, 0) + bits[17:]
            # Check for special values in the remainder of the combination:
            more = bits[3:6]
            if more == (1, 0, 0) or more == (1, 0, 1):
                spelling = "Infinity"
            elif more == (1, 1, 0):
                spelling = "NaN (quiet)"
            elif more == (1, 1, 1):
                spelling = "NaN (signaling)"

        coeff = mergebits(coeff)
        exponent = mergebits(exponent)
        if spelling is None:
            spelling = str(coeff)
            e = exponent - 6176
            if e == 0:
                pass
            elif e < 0:
                spelling = spelling.zfill(abs(e))
                split = len(spelling) + e
                w, fr = spelling[:split], spelling[split:]
                spelling = f"{w}.{fr}"
            else:
                spelling = spelling + "0" * e

        if sign:
            spelling = f"-{spelling}"

        # The "combination" bits
        combination = mergebits(bits[1:18])
        return Decimal128Value(sign, combination, exponent, coeff, spelling)

    @override
    def get_children(self) -> Iterable[ChildItem]:
        yield "sign", self.value.sign
        yield "combination", self.value.combination, lldb.eFormatBinary, None
        yield "exponent (biased)", self.value.exponent
        yield "exponent (actual)", self.value.exponent - 6176
        yield "significand", str(self.value.significand)
        yield "value", self.value.spelling


class MaxKeyDisplay(NullDisplay):
    """The display type for BSON's 'max key' type"""

    __typename__ = "__bson_maxkey__"
    __summary_str__ = "max key"


class MinKeyDisplay(NullDisplay):
    """The display type for BSON's 'min key' type"""

    __typename__ = "__bson_minkey__"
    __summary_str__ = "min key"


class BSONTInfo(NamedTuple):
    """Information about a bson_t object"""

    addr: int
    "The address of the pointer to the beginning of the BSON data managed by this object"
    size: int
    "The size of the BSON data managed/referenced by this object"
    flags: int
    "Flags of the bson_t object"


class BSONTError(NamedTuple):
    """Represents an error while reading a bson_t object"""

    reason: str
    "A description of the error that ocurred"


class BSONTDisplay(SyntheticDisplayBase["BSONTInfo | BSONTError"]):
    """
    Implements inspection logic for bson_t
    """

    __typename__ = "bson_t"

    @classmethod
    @override
    def __parse__(cls, value: SBValue) -> BSONTInfo | BSONTError:
        # Dereference pointers:
        if value.TypeIsPointerType():
            value = value.deref

        # Unwrap the wrapped object
        value = value.GetNonSyntheticValue()

        dat = value.data
        # We know the location of the flags and size without debug info:
        err = SBError()
        flags = dat.GetUnsignedInt32(err, 0)
        if err.fail:
            return BSONTError(f"Failed to read memory at 0x{value.load_addr:x}: {err.description}")
        length = dat.GetUnsignedInt32(err, 4)
        if err.fail:
            return BSONTError(f"Failed to read memory at 0x{value.load_addr:x}: {err.description}")

        # Check bogus values:
        MAX_SIZE = 16 * 1024 * 1024
        ALL_FLAGS = (1 << 6) - 1
        if flags & ~ALL_FLAGS or length < 5 or length > MAX_SIZE:
            return BSONTError(f"bson_t appears uninitialized/invalid [a] {flags=} {length=}")

        is_inline = bool(flags & 1)

        if is_inline:
            # Inline objects may only occupy 120 bytes, at most
            if length > 120:
                return BSONTError("bson_t appears uninitialized/invalid [b]")
            # Look for debug info for the inline impl
            inline_t = value.target.FindFirstType("bson_impl_inline_t")
            if inline_t:
                as_inline = value.Cast(inline_t)
                ptr = as_inline.GetChildMemberWithName("data").load_addr
            else:
                # No debug info? Guess its location as the default
                ptr = value.load_addr + 4 + 4
            if not err.success:
                return BSONTError(f"Failed to read inline bson_t data: {err}")
            return BSONTInfo(ptr, length, flags)

        # Look for impl_alloc_t
        alloc_t = value.target.FindFirstType("bson_impl_alloc_t")
        if alloc_t:
            alloc = value.Cast(alloc_t)
            # Walk to the buffer for this value:
            offset = alloc.GetChildMemberWithName("offset").unsigned
            buf = alloc.GetChildMemberWithName("buf").deref.deref
            ptr = buf.load_addr + offset
            return BSONTInfo(ptr, length, flags)

        # No debug info, we have to calc the offset ourself
        u8_t = value.target.GetBasicType(lldb.eBasicTypeUnsignedChar)
        u8ptr_t = u8_t.GetPointerType()
        ptr_size = u8ptr_t.size
        u32_size = 4
        buf_off = (u32_size * 2) + ptr_size + u32_size
        if u32_size < ptr_size:
            # Adjust for alignment
            buf_off += u32_size
        offset_off = buf_off + (ptr_size * 2)
        offset = dat.GetUnsignedInt32(err, offset_off)
        if not err.success:
            return BSONTError(f"Failed to read offset of buffer: {err}")
        bufptr = value.CreateChildAtOffset("buf", buf_off, u8ptr_t.GetPointerType()).deref
        if not bufptr.error.success:
            return BSONTError(f"Failed to read the alloc buf: {bufptr.error} {offset=} {buf_off=}")
        ptr = bufptr.data.GetUnsignedInt64(err, 0)
        assert err.success, err

        u32_t = value.target.FindFirstType("uint32_t")
        addr = SBAddress()
        addr.SetLoadAddress(ptr, value.target)

        u32 = value.target.CreateValueFromAddress("tmp", addr, u32_t)
        assert u32.error.success, u32
        if u32.unsigned != length or length < 5:
            return BSONTError(f"bson_t appears uninitialized/invalid [c] {flags=} {length=} {u32.unsigned=}")
        return BSONTInfo(ptr, length, flags)

    @override
    def get_children(self) -> Iterable[ChildItem]:
        val = self.value
        if isinstance(val, BSONTError):
            yield "[error]", val.reason
            return

        # Imbue the flags with the possible debug info to give it a nice rendering
        flags_t = self.sbvalue.target.FindFirstType("bson_flags_t")
        if flags_t.IsValid():
            yield "flags", val.flags, None, flags_t
        else:
            yield "flags", val.flags
        yield "data size", val.size
        ptr_t = self.sbvalue.target.GetBasicType(lldb.eBasicTypeVoid).GetPointerType()
        yield "data address", val.addr, lldb.eFormatPointer, ptr_t

        # Generate the __bson_document_xxx__ that will allow walking the document:
        doc_t = DocumentDisplay.__get_sbtype__(self.sbvalue.frame, val.addr)
        yield lambda: checked(self.sbvalue.synthetic_child_from_address("[content]", val.addr, doc_t))


def checked(val: SBValue) -> SBValue:
    """Assert that ``val`` is valid. Returns ``val``"""
    assert val.error.success, f"{val=} {val.error=}"
    return val


def read_i32le(dat: bytes) -> int:
    """Read a 32-bit integer from the given data."""
    # Truncate before the read:
    buf = dat[0:4]
    return struct.unpack("<i", buf)[0]


_types_cache: dict[tuple[int, str], SBType] = {}
"The cache of generated types (for generate_or_get_type)"


def generate_or_get_type(expr_prefix: str, frame: SBFrame) -> SBType:
    """
    This is the big magic in this module: LLDB doesn't have a "simple" API to
    generate new datatypes (which we need in order to inspect element contents!),
    so we instead generate them on-the-fly using the expression evaluator.

    If we evaluate an expression like:

        struct some_struct { int a, b, c; };
        some_struct x;
        x

    LLDB will give us a handle to ``x``. We can then ask for the type of ``x``,
    which gives us a handle to `some_struct`, from which we can now synthesize
    new values and use as if it were part of the target's debug information all
    along.

    This needs to be done carefully, since it is an expensive operation and
    we do not want to declare the same type more than once. For this reason,
    we cache.
    """
    # The cache key
    cachekey = frame.thread.process.id, expr_prefix
    existing = _types_cache.get(cachekey)
    if existing is not None:
        # We've already generated this type before
        return existing
    # Create a new temporary object. Give it a unique name to prevent it from
    # colliding with any possible temporaries we may have generated previously.
    hash = hashlib.md5(expr_prefix.encode()).hexdigest()
    varname = f"__bson_lldb_tmp_{hash}"
    full_expr = f"{expr_prefix} {varname}; {varname}"
    tmp = frame.EvaluateExpression(full_expr)
    existing = tmp.type
    _types_cache[cachekey] = existing
    return existing


class BSONType(enum.Enum):
    """The values for each bson element type tag"""

    EOD = 0
    Double = 0x1
    UTF8 = 0x2
    Document = 0x3
    Array = 0x4
    Binary = 0x5
    Undefined = 0x6
    ObjectID = 0x7
    Bool = 0x8
    Datetime = 0x9
    Null = 0xA
    Regex = 0xB
    DBPointer = 0xC
    Code = 0xD
    Symbol = 0xE
    CodeWithScope = 0xF
    Int32 = 0x10
    Timestamp = 0x11
    Int64 = 0x12
    Decimal128 = 0x13
    MinKey = 255
    MaxKey = 127


class _BSONWalker:
    """
    This implement document traversal in a Python expression evaluator. It
    uses __getattr__ and __getitem__ to provide simple semantics for accessing
    subelements, and uses the matmul infix operator ``@`` to resolve the path
    against a given SBValue that refers to a ``bson_t``
    """

    def __init__(self, path: Iterable[str | int] = ()) -> None:
        self._path = tuple(path)
        "The path that we will resolve when '@' is applied."

    def __rmatmul__(self, lhs: Any) -> SBValue | str:
        """Implement the actual path resolution. This is the '@' operator."""
        if isinstance(lhs, str):
            # Evaluate the left-hand string as an expression within the target
            target = lldb.debugger.GetSelectedTarget()
            if target is None:
                raise RuntimeError("Not attached to a debug target")
            frame = target.process.selected_thread.frames[0]
            lhs = frame.EvaluateExpression(lhs)
        val: SBValue
        if hasattr(lhs.__class__, "unwrap"):
            # CodeLLDB gives us a wrapper around SBValue, but we want the unwrapped
            # version:
            val = lhs.__class__.unwrap(lhs)
        else:
            val = lhs

        if val.TypeIsPointerType():
            # Dereference pointers
            val = val.deref

        # Drop the synthetic wrapper that was wrapped around bson_t
        val = val.GetNonSyntheticValue()

        # Parse it:
        as_bson = BSONTDisplay.__parse__(val)
        if isinstance(as_bson, BSONTError):
            raise ValueError(as_bson.reason)

        # Create the synthetic __bson_document_xxx__ object for this doc
        doc_t = DocumentDisplay.__get_sbtype__(val.frame, as_bson.addr)
        # Obtain a value reference to the document data:
        retval = val.CreateValueFromAddress("[root]", as_bson.addr, doc_t)

        # Now resolve the path:
        for part in self._path:
            if isinstance(part, str):
                # Access via ``p['foo']`` or ``p.foo``, requires our current node
                # to be a document:
                if not retval.type.name.startswith("__bson_document_"):
                    raise AttributeError(
                        f'Element of type {retval.type.name} cannot be accessed as a document (looking for element "{part}")'
                    )
                # We are accessing the synthetic children generated by DocumentDisplay.
                # The keys are of the following format based on the element keys in the document:
                want_child_name = f"['{part}']"
            else:
                # Access via indexing ``p[42]``, requires an array
                if not retval.type.name.startswith("__bson_array_"):
                    raise AttributeError(
                        f"Element of type {retval.type.name} cannot be accessed as an array (looking for element {part})"
                    )
                # Array keys are bracketed, but not quoted
                want_child_name = f"[{part}]"
            # Find all children that match the key (usually only one)
            matching = (c for c in retval.children if c.name == want_child_name)
            # Get it:
            got = next(iter(matching), None)
            if got is None:
                # Didn't get it...
                if isinstance(part, str):
                    raise KeyError(f'Document has no element "{part}"')
                else:
                    raise IndexError(f"Array index [{part}] is out-of-bounds")
            # Set this as our current node, which we may step in further, or
            # we may be done
            retval = got

        return retval

    def __getattr__(self, key: str) -> _BSONWalker:
        """Generate a new traversal for the given key"""
        return _BSONWalker(self._path + (key,))

    def __getitem__(self, key: str | int) -> _BSONWalker:
        """Generate a new traversal for the given key or index"""
        return _BSONWalker(self._path + (key,))


class _MemoryCache:
    """
    Reading process memory is an extremely slow operation, and we don't want to
    re-read memory multiple times. Because parsing a document's elements requires
    reading the entire document into memory, we are guaranteed that the top-level
    DocumentDisplay object will perform a ReadMemory() of the entire object, and
    thus all child elements can reuse that buffer when they parse and render
    their own contents.

    This class implements that cache of memory segments
    """

    def __init__(self):
        self._segments: dict[int, bytes] = {}
        "Segments of memory keyed by the base address of the read operation"

    def get_cached(self, addr: int) -> bytes:
        """
        Find and return a chunk of memory that was previously read from the
        process. The ``addr`` must be an address within a BSON document that
        was previously read by a DocumentDisplay.

        The returned buffer begins at exactly ``addr``, and contains the full
        remainder of the cached segment, which may be much longer than the
        caller actually needs. We don't need additional granularity for our
        purposes, though.
        """
        segment = self.segment_containing(addr)
        if not segment:
            # Memory does not exist?
            print(f"lldb_bson: Note: Attempted read of uncached address 0x{addr:x}")
            print("".join(traceback.format_stack()))
            return b"\0" * 512
        base_addr, data = segment
        inner_offset = addr - base_addr
        return data[inner_offset:]

    def segment_containing(self, addr: int) -> tuple[int, bytes] | None:
        """Find the segment that contains ``addr``, or returns None"""
        for base_addr, data in self._segments.items():
            end_addr = base_addr + len(data)
            if addr > end_addr or addr < base_addr:
                continue
            return base_addr, data
        return None

    def read(self, val: SBValue) -> tuple[int, bytes]:
        """
        Read the memory segment referenced by the given SBValue
        """
        return self.read_at(val.process, val.load_addr, val.size)

    def read_at(self, proc: SBProcess, addr: int, size: int) -> tuple[int, bytes]:
        """
        Read from ``proc`` a chunk of memory beginning at ``addr`` and continuing
        for ``size`` bytes.
        """
        err = SBError()
        buf = proc.ReadMemory(addr, size, err)
        if err.fail:
            raise LookupError(err.description)
        self._segments[addr] = buf
        return addr, buf


memcache = _MemoryCache()
"A module-wide memory segment cache."



# Copyright 2009-present MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import MutableMapping
from evergreen_config_generator import ConfigObject

from . import Value, ValueSequence


class TaskGroup(ConfigObject):
    def __init__(self, name: str):
        self._task_group_name = name
        self.setup_group: ValueSequence | None = None
        self.teardown_group: ValueSequence | None = None
        self.setup_task: str | None = None
        self.teardown_task: str | None = None
        self.max_hosts: int | None = None
        self.timeout: int | None = None
        self.setup_group_can_fail_task: bool | None = None
        self.setup_group_timeout_secs: int | None = None
        self.share_processes: bool | None = None
        self.tasks: ValueSequence | None = None

    @property
    def name(self) -> str:
        return self._task_group_name

    def to_dict(self) -> Value:
        v = super().to_dict()
        assert isinstance(v, MutableMapping)
        # See possible TaskGroup attributes from the Evergreen wiki:
        # https://github.com/evergreen-ci/evergreen/wiki/Project-Configuration-Files#task-groups
        attrs = [
            "setup_group",
            "teardown_group",
            "setup_task",
            "teardown_task",
            "max_hosts",
            "timeout",
            "setup_group_can_fail_task",
            "setup_group_timeout_secs",
            "share_processes",
            "tasks",
        ]

        for i in attrs:
            if getattr(self, i, None):
                v[i] = getattr(self, i)
        return v



from itertools import product

from shrub.v3.evg_command import expansions_update
from shrub.v3.evg_command import KeyValueParam
from shrub.v3.evg_task import EvgTask, EvgTaskDependency

from config_generator.etc.distros import find_small_distro
from config_generator.etc.distros import make_distro_str
from config_generator.etc.distros import to_cc

from config_generator.components.funcs.bootstrap_mongo_orchestration import BootstrapMongoOrchestration
from config_generator.components.funcs.fetch_build import FetchBuild
from config_generator.components.funcs.fetch_det import FetchDET
from config_generator.components.funcs.run_simple_http_server import RunSimpleHTTPServer
from config_generator.components.funcs.run_mock_kms_servers import RunMockKMSServers
from config_generator.components.funcs.run_tests import RunTests


def generate_test_tasks(SSL, TAG, MATRIX, MORE_COMPILE_TAGS=None, MORE_TEST_TAGS=None, MORE_VARS=None):
    res = []

    TOPOLOGY_TO_STR = {
        'server': 'server',
        'replica': 'replica_set',
        'sharded': 'sharded_cluster',
    }

    MORE_COMPILE_TAGS = MORE_COMPILE_TAGS if MORE_COMPILE_TAGS else []
    MORE_TEST_TAGS = MORE_TEST_TAGS if MORE_TEST_TAGS else []
    MORE_VARS = MORE_VARS if MORE_VARS else {}

    for distro_name, compiler, arch, sasl, auths, topologies, server_vers in MATRIX:
        tags = [
            TAG, 'test', distro_name, compiler, f'sasl-{sasl}'
        ] + MORE_COMPILE_TAGS

        test_distro = find_small_distro(distro_name)

        compile_vars = []
        compile_vars.append(KeyValueParam(key='CC', value=to_cc(compiler)))

        if arch:
            tags.append(arch)
            compile_vars.append(KeyValueParam(key='MARCH', value=arch))

        distro_str = make_distro_str(distro_name, compiler, arch)

        base_task_name = f'sasl-{sasl}-{SSL}-{distro_str}'

        for tag in MORE_COMPILE_TAGS:
            base_task_name = f'{tag}-{base_task_name}'

        compile_task_name = f'{base_task_name}-compile'

        for auth, topology, server_ver in product(auths, topologies, server_vers):
            test_tags = tags + [auth, topology, server_ver] + MORE_TEST_TAGS
            # Do not add `nossl` tag to prevent being selected by legacy config variants.
            # Remove the `if` when CDRIVER-4571 is resolved.
            if SSL != 'nossl':
                test_tags += [SSL]
            test_task_name = f'{base_task_name}-test-{server_ver}-{topology}-{auth}'

            for tag in MORE_TEST_TAGS:
                test_task_name += f'-{tag}'

            test_commands = []
            test_commands.append(FetchBuild.call(build_name=compile_task_name))

            updates = compile_vars + [
                KeyValueParam(key='AUTH', value=auth),
                KeyValueParam(key='MONGODB_VERSION', value=server_ver),
                KeyValueParam(key='TOPOLOGY', value=TOPOLOGY_TO_STR[topology]),
                KeyValueParam(key='SSL', value=SSL),
            ]

            if 'cse' in MORE_COMPILE_TAGS:
                updates.append(
                    KeyValueParam(key='CLIENT_SIDE_ENCRYPTION', value='on')
                )

            for key, value in MORE_VARS.items():
                updates.append(KeyValueParam(key=key, value=value))

            test_commands.append(expansions_update(updates=updates))
            test_commands.append(FetchDET.call())
            test_commands.append(BootstrapMongoOrchestration.call())
            test_commands.append(RunSimpleHTTPServer.call())

            if 'cse' in MORE_COMPILE_TAGS:
                test_commands.append(RunMockKMSServers.call())

            test_commands.append(RunTests.call())

            res.append(
                EvgTask(
                    name=test_task_name,
                    run_on=test_distro.name,
                    tags=test_tags,
                    depends_on=[EvgTaskDependency(name=compile_task_name)],
                    commands=test_commands,
                )
            )

    return res



from shrub.v3.evg_command import EvgCommandType

from config_generator.etc.function import Function
from config_generator.etc.utils import bash_exec


class RunSimpleHTTPServer(Function):
    name = 'run-simple-http-server'
    command_type = EvgCommandType.SETUP
    commands = [
        bash_exec(
            command_type=command_type,
            background=True,
            working_dir='mongoc',
            script='''\
                set -o errexit
                echo "Starting simple HTTP server..."
                command -V "${PYTHON3_BINARY}" >/dev/null
                "${PYTHON3_BINARY}" .evergreen/scripts/simple_http_server.py
                echo "Starting simple HTTP server... done."
            ''',
        ),
    ]

    @classmethod
    def call(cls, **kwargs):
        return cls.default_call(**kwargs)


def functions():
    return RunSimpleHTTPServer.defn()



from shrub.v3.evg_command import EvgCommandType
from shrub.v3.evg_command import s3_put
from shrub.v3.evg_task import EvgTask

from config_generator.components.funcs.find_cmake_latest import FindCMakeLatest

from config_generator.etc.function import Function
from config_generator.etc.function import merge_defns
from config_generator.etc.utils import bash_exec


class MakeDocs(Function):
    name = "make-docs"
    commands = [
        bash_exec(
            command_type=EvgCommandType.TEST,
            working_dir="mongoc",
            include_expansions_in_env=["distro_id"],
            script="""\
                set -o errexit
                ./tools/poetry.sh install --with=docs
                # See SphinxBuild.cmake for EVG_DOCS_BUILD reasoning
                ./tools/poetry.sh run env EVG_DOCS_BUILD=1 .evergreen/scripts/build-docs.sh
                """,
        ),
    ]

    @classmethod
    def call(cls, **kwargs):
        return cls.default_call(**kwargs)


class UploadDocs(Function):
    name = "upload-docs"
    commands = [
        bash_exec(
            working_dir="mongoc/_build/for-docs/src/libbson",
            env={
                "AWS_ACCESS_KEY_ID": "${aws_key}",
                "AWS_SECRET_ACCESS_KEY": "${aws_secret}",
            },
            script="aws s3 cp doc/html s3://mciuploads/mongo-c-driver/docs/libbson/${CURRENT_VERSION} --quiet --recursive --acl public-read --region us-east-1",
        ),
        s3_put(
            aws_key="${aws_key}",
            aws_secret="${aws_secret}",
            bucket="mciuploads",
            content_type="text/html",
            display_name="libbson docs",
            local_file="mongoc/_build/for-docs/src/libbson/doc/html/index.html",
            permissions="public-read",
            remote_file="mongo-c-driver/docs/libbson/${CURRENT_VERSION}/index.html",
        ),
        bash_exec(
            working_dir="mongoc/_build/for-docs/src/libmongoc",
            env={
                "AWS_ACCESS_KEY_ID": "${aws_key}",
                "AWS_SECRET_ACCESS_KEY": "${aws_secret}",
            },
            script="aws s3 cp doc/html s3://mciuploads/mongo-c-driver/docs/libmongoc/${CURRENT_VERSION} --quiet --recursive --acl public-read --region us-east-1",
        ),
        s3_put(
            aws_key="${aws_key}",
            aws_secret="${aws_secret}",
            bucket="mciuploads",
            content_type="text/html",
            display_name="libmongoc docs",
            local_file="mongoc/_build/for-docs/src/libmongoc/doc/html/index.html",
            permissions="public-read",
            remote_file="mongo-c-driver/docs/libmongoc/${CURRENT_VERSION}/index.html",
        ),
    ]

    @classmethod
    def call(cls, **kwargs):
        return cls.default_call(**kwargs)


class UploadManPages(Function):
    name = "upload-man-pages"
    commands = [
        bash_exec(
            working_dir="mongoc",
            silent=True,
            script="""\
                set -o errexit
                # Get "aha", the ANSI HTML Adapter.
                git clone --depth 1 https://github.com/theZiz/aha.git aha-repo
                pushd aha-repo
                make
                popd # aha-repo
                mv aha-repo/aha .
                .evergreen/scripts/man-pages-to-html.sh libbson _build/for-docs/src/libbson/doc/man > bson-man-pages.html
                .evergreen/scripts/man-pages-to-html.sh libmongoc _build/for-docs/src/libmongoc/doc/man > mongoc-man-pages.html
                """,
        ),
        s3_put(
            aws_key="${aws_key}",
            aws_secret="${aws_secret}",
            bucket="mciuploads",
            content_type="text/html",
            display_name="libbson man pages",
            local_file="mongoc/bson-man-pages.html",
            permissions="public-read",
            remote_file="mongo-c-driver/man-pages/libbson/${CURRENT_VERSION}/index.html",
        ),
        s3_put(
            aws_key="${aws_key}",
            aws_secret="${aws_secret}",
            bucket="mciuploads",
            content_type="text/html",
            display_name="libmongoc man pages",
            local_file="mongoc/mongoc-man-pages.html",
            permissions="public-read",
            remote_file="mongo-c-driver/man-pages/libmongoc/${CURRENT_VERSION}/index.html",
        ),
    ]

    @classmethod
    def call(cls, **kwargs):
        return cls.default_call(**kwargs)


def functions():
    return merge_defns(
        MakeDocs.defn(),
        UploadDocs.defn(),
        UploadManPages.defn(),
    )


def tasks():
    return [
        EvgTask(
            name="make-docs",
            commands=[
                FindCMakeLatest.call(),
                MakeDocs.call(),
                UploadDocs.call(),
                UploadManPages.call(),
            ],
        )
    ]



from shrub.v3.evg_command import EvgCommandType

from config_generator.etc.function import Function
from config_generator.etc.utils import bash_exec


class RunMockKMSServers(Function):
    name = 'run-mock-kms-servers'
    command_type = EvgCommandType.SETUP
    commands = [
        # This command ensures future invocations of activate-kmstlsvenv.sh conducted in
        # parallel do not race to setup a venv environment; it has already been prepared.
        # This primarily addresses the situation where the "run tests" and "run-mock-kms-servers"
        # functions invoke 'activate-kmstlsvenv.sh' simultaneously.
        # TODO: remove this function along with the "run-mock-kms-servers" function.
        bash_exec(
            command_type=command_type,
            working_dir='drivers-evergreen-tools/.evergreen/csfle',
            script='''\
                set -o errexit
                echo "Preparing KMS TLS venv environment..."
                # TODO: remove this function along with the "run kms servers" function.
                if [[ "$OSTYPE" =~ cygwin && ! -d kmstlsvenv ]]; then
                    # Avoid using Python 3.10 on Windows due to incompatible cipher suites.
                    # See CDRIVER-4530.
                    . ../venv-utils.sh
                    venvcreate "C:/python/Python39/python.exe" kmstlsvenv || # windows-2017
                    venvcreate "C:/python/Python38/python.exe" kmstlsvenv    # windows-2015
                    python -m pip install --upgrade boto3~=1.19 pykmip~=0.10.0 "sqlalchemy<2.0.0"
                    deactivate
                else
                    . ./activate-kmstlsvenv.sh
                    deactivate
                fi
                echo "Preparing KMS TLS venv environment... done."
            ''',
        ),
        bash_exec(
            command_type=command_type,
            background=True,
            working_dir='drivers-evergreen-tools/.evergreen/csfle',
            script='''\
                set -o errexit
                echo "Starting mock KMS TLS servers..."
                . ./activate-kmstlsvenv.sh
                python -u kms_http_server.py --ca_file ../x509gen/ca.pem --cert_file ../x509gen/server.pem --port 8999 &
                python -u kms_http_server.py --ca_file ../x509gen/ca.pem --cert_file ../x509gen/expired.pem --port 9000 &
                python -u kms_http_server.py --ca_file ../x509gen/ca.pem --cert_file ../x509gen/wrong-host.pem --port 9001 &
                python -u kms_http_server.py --ca_file ../x509gen/ca.pem --cert_file ../x509gen/server.pem --require_client_cert --port 9002 &
                python -u kms_failpoint_server.py --port 9003 &
                python -u kms_kmip_server.py &
                deactivate
                echo "Starting mock KMS TLS servers... done."
            ''',
        ),
    ]

    @classmethod
    def call(cls, **kwargs):
        return cls.default_call(**kwargs)


def functions():
    return RunMockKMSServers.defn()



from shrub.v3.evg_command import archive_targz_extract
from shrub.v3.evg_command import EvgCommandType
from shrub.v3.evg_command import s3_get

from config_generator.etc.function import Function
from config_generator.etc.utils import bash_exec


class FetchBuild(Function):
    name = 'fetch-build'
    command_type = EvgCommandType.SETUP
    commands = [
        bash_exec(command_type=command_type, script='rm -rf mongoc'),
        s3_get(
            command_type=command_type,
            aws_key='${aws_key}',
            aws_secret='${aws_secret}',
            bucket='mciuploads',
            local_file='build.tar.gz',
            remote_file='mongo-c-driver/${build_variant}/${revision}/${BUILD_NAME}/${build_id}.tar.gz',
        ),
        archive_targz_extract(path='build.tar.gz', destination='mongoc'),
        # Scripts may not be executable on Windows.
        bash_exec(
            command_type=command_type,
            working_dir='mongoc',
            script='''\
                for file in $(find .evergreen/scripts -type f); do
                    chmod +rx "$file" || exit
                done
            '''
        ),
    ]

    @classmethod
    def call(cls, build_name, vars=None, **kwargs):
        if vars is None:
            vars = {}
        vars.update({'BUILD_NAME': build_name})
        return cls.default_call(vars=vars, **kwargs)


def functions():
    return FetchBuild.defn()



from shrub.v3.evg_command import EvgCommandType
from shrub.v3.evg_command import expansions_update
from shrub.v3.evg_command import git_get_project

from config_generator.etc.function import Function
from config_generator.etc.utils import bash_exec


class FetchSource(Function):
    name = 'fetch-source'
    command_type = EvgCommandType.SETUP
    commands = [
        git_get_project(command_type=command_type, directory='mongoc'),
        bash_exec(
            command_type=command_type,
            working_dir='mongoc',
            script='''\
                set -o errexit
                set -o pipefail
                if [ -n "${github_pr_number}" -o "${is_patch}" = "true" ]; then
                    VERSION=patch-${version_id}
                else
                    VERSION=latest
                fi
                echo "CURRENT_VERSION: $VERSION" > expansion.yml
            '''
        ),
        expansions_update(command_type=command_type,
                          file='mongoc/expansion.yml'),
        # Scripts may not be executable on Windows.
        bash_exec(
            command_type=EvgCommandType.SETUP,
            working_dir='mongoc',
            script='''\
                for file in $(find .evergreen/scripts -type f); do
                    chmod +rx "$file" || exit
                done
            '''
        ),
    ]

    @classmethod
    def call(cls, **kwargs):
        return cls.default_call(**kwargs)


def functions():
    return FetchSource.defn()



from shrub.v3.evg_command import EvgCommandType

from config_generator.etc.function import Function
from config_generator.etc.utils import bash_exec


class FindCMakeLatest(Function):
    '''
    Call `find_cmake_latest` in an attempt to download-and-build the latest
    CMake version as a Setup task with `retry_on_failure: true` prior to
    subsequent use of `find-cmake-latest.sh` by compile and build scripts.
    '''

    name = 'find-cmake-latest'
    command_type = EvgCommandType.SETUP
    commands = [
        bash_exec(
            command_type=command_type,
            retry_on_failure=True,
            working_dir='mongoc',
            script='. .evergreen/scripts/find-cmake-latest.sh && find_cmake_latest'
        ),
    ]

    @classmethod
    def call(cls, **kwargs):
        return cls.default_call(**kwargs)


def functions():
    return FindCMakeLatest.defn()



#!/usr/bin/env python3

"""
Python script to interface as a mock OCSP responder.
"""

import argparse
import logging
import sys
import os

sys.path.append(os.path.join(os.getcwd() ,'src', 'third_party', 'mock_ocsp_responder'))

import mock_ocsp_responder

def main():
    """Main entry point"""
    parser = argparse.ArgumentParser(description="MongoDB Mock OCSP Responder.")

    parser.add_argument('-p', '--port', type=int, default=8080, help="Port to listen on")

    parser.add_argument('--ca_file', type=str, required=True, help="CA file for OCSP responder")

    parser.add_argument('-v', '--verbose', action='count', help="Enable verbose tracing")

    parser.add_argument('--ocsp_responder_cert', type=str, required=True, help="OCSP Responder Certificate")

    parser.add_argument('--ocsp_responder_key', type=str, required=True, help="OCSP Responder Keyfile")

    parser.add_argument('--fault', choices=[mock_ocsp_responder.FAULT_REVOKED, mock_ocsp_responder.FAULT_UNKNOWN, None], default=None, type=str, help="Specify a specific fault to test")

    parser.add_argument('--next_update_seconds', type=int, default=32400, help="Specify how long the OCSP response should be valid for")

    args = parser.parse_args()
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)

    print('Initializing OCSP Responder')
    mock_ocsp_responder.init_responder(issuer_cert=args.ca_file, responder_cert=args.ocsp_responder_cert, responder_key=args.ocsp_responder_key, fault=args.fault, next_update_seconds=args.next_update_seconds)

    if args.verbose:
        mock_ocsp_responder.init(args.port, debug=True)
    else:
        mock_ocsp_responder.init(args.port)

    print('Mock OCSP Responder is running on port %s' % (str(args.port)))

if __name__ == '__main__':
    main()



from config_generator.etc.distros import find_small_distro
from config_generator.etc.function import Function, merge_defns
from config_generator.etc.utils import bash_exec

from shrub.v3.evg_build_variant import BuildVariant
from shrub.v3.evg_command import BuiltInCommand, EvgCommandType, expansions_update, s3_put
from shrub.v3.evg_task import EvgTask, EvgTaskRef

from pydantic import ConfigDict
from typing import Optional


TAG = 'sbom'


class CustomCommand(BuiltInCommand):
    command: str
    model_config = ConfigDict(arbitrary_types_allowed=True)


def ec2_assume_role(
    role_arn: Optional[str] = None,
    policy: Optional[str] = None,
    duration_seconds: Optional[int] = None,
    command_type: Optional[EvgCommandType] = None,
) -> CustomCommand:
    return CustomCommand(
        command="ec2.assume_role",
        params={
            "role_arn": role_arn,
            "policy": policy,
            "duration_seconds": duration_seconds,
        },
        type=command_type,
    )


class SBOM(Function):
    name = 'sbom'
    commands = [
        ec2_assume_role(
            command_type=EvgCommandType.SETUP,
            role_arn='${kondukto_role_arn}',
        ),
        bash_exec(
            command_type=EvgCommandType.SETUP,
            include_expansions_in_env=[
                'AWS_ACCESS_KEY_ID',
                'AWS_SECRET_ACCESS_KEY',
                'AWS_SESSION_TOKEN',
            ],
            script='''\
                set -o errexit
                set -o pipefail
                kondukto_token="$(aws secretsmanager get-secret-value --secret-id "kondukto-token" --region "us-east-1" --query 'SecretString' --output text)"
                printf "KONDUKTO_TOKEN: %s\\n" "$kondukto_token" >|expansions.kondukto.yml
            ''',
        ),
        expansions_update(
            command_type=EvgCommandType.SETUP,
            file='expansions.kondukto.yml',
        ),
        bash_exec(
            command_type=EvgCommandType.TEST,
            working_dir='mongoc',
            include_expansions_in_env=[
                'artifactory_password',
                'artifactory_username',
                'branch_name',
                'KONDUKTO_TOKEN',
            ],
            script='.evergreen/scripts/sbom.sh',
        ),
        s3_put(
            command_type=EvgCommandType.TEST,
            aws_key='${aws_key}',
            aws_secret='${aws_secret}',
            bucket='mciuploads',
            content_type='application/json',
            display_name='Augmented SBOM',
            local_file='mongoc/augmented-sbom.json',
            permissions='public-read',
            remote_file='mongo-c-driver/${build_variant}/${revision}/${version_id}/${build_id}/sbom/augmented-sbom.json',
        ),
    ]

    @classmethod
    def call(cls, **kwargs):
        return cls.default_call(**kwargs)


def functions():
    return merge_defns(
        SBOM.defn(),
    )


def tasks():
    distro_name = 'rhel80'
    distro = find_small_distro(distro_name)

    yield EvgTask(
        name='sbom',
        tags=[TAG, distro_name],
        run_on=distro.name,
        commands=[
            SBOM.call(),
        ],
    )


def variants():
    return [
        BuildVariant(
            name=TAG,
            display_name='SBOM',
            tasks=[EvgTaskRef(name=f'.{TAG}')],
        ),
    ]



from shrub.v3.evg_task import EvgTask

from config_generator.etc.distros import find_large_distro
from config_generator.etc.distros import make_distro_str
from config_generator.etc.distros import to_cc

from config_generator.components.funcs.find_cmake_latest import FindCMakeLatest
from config_generator.components.funcs.upload_build import UploadBuild


def generate_compile_tasks(SSL, TAG, SASL_TO_FUNC, MATRIX, MORE_TAGS=None, MORE_VARS=None):
    res = []

    MORE_TAGS = MORE_TAGS if MORE_TAGS else []
    MORE_VARS = MORE_VARS if MORE_VARS else {}

    for distro_name, compiler, arch, sasls, in MATRIX:
        tags = [TAG, 'compile', distro_name, compiler] + MORE_TAGS

        distro = find_large_distro(distro_name)

        compile_vars = None
        compile_vars = {'CC': to_cc(compiler)}

        if arch:
            tags.append(arch)
            compile_vars.update({'MARCH': arch})

        compile_vars.update(MORE_VARS)

        distro_str = make_distro_str(distro_name, compiler, arch)

        for sasl in sasls:
            task_name = f'sasl-{sasl}-{SSL}-{distro_str}-compile'

            for tag in MORE_TAGS:
                task_name = f'{tag}-{task_name}'

            commands = []
            commands.append(FindCMakeLatest.call())
            commands.append(SASL_TO_FUNC[sasl].call(vars=compile_vars))
            commands.append(UploadBuild.call())

            res.append(
                EvgTask(
                    name=task_name,
                    run_on=distro.name,
                    tags=tags + [f'sasl-{sasl}'],
                    commands=commands,
                )
            )

    return res



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

from pathlib import Path
import re
from typing import Callable

from sphinx import addnodes
from sphinx.application import Sphinx
from sphinx.environment import BuildEnvironment

THIS_FILE = Path(__file__).resolve()
THIS_DIR = THIS_FILE.parent
REPO_ROOT = THIS_DIR.parent.parent

project = "MongoDB C Driver Development"
copyright = "2009-present, MongoDB, Inc."
author = "MongoDB, Inc"
release = (REPO_ROOT / "VERSION_CURRENT").read_text().strip()

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = []
templates_path = []
exclude_patterns = []
default_role = "any"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "nature"
pygments_style = "sphinx"
html_static_path = []

rst_prolog = rf"""
.. role:: bash(code)
    :language: bash
"""


def annotator(
    annot: str,
) -> Callable[[BuildEnvironment, str, addnodes.desc_signature], str]:
    """
    Create a parse_node function that adds a parenthesized annotation to an object signature.
    """

    def parse_node(
        env: BuildEnvironment, sig: str, signode: addnodes.desc_signature
    ) -> str:
        signode += addnodes.desc_name(sig, sig)
        signode += addnodes.desc_sig_space()
        signode += addnodes.desc_annotation("", f"({annot})")
        return sig

    return parse_node


def parse_earthly_artifact(
    env: BuildEnvironment, sig: str, signode: addnodes.desc_signature
) -> str:
    """
    Parse and render the signature of an '.. earthly-artifact::' signature"""
    mat = re.match(r"(?P<target>\+.+?)(?P<path>/.*)$", sig)
    if not mat:
        raise RuntimeError(
            f"Invalid earthly-artifact signature: {sig!r} (expected “+<target>/<path> string)"
        )
    signode += addnodes.desc_addname(mat["target"], mat["target"])
    signode += addnodes.desc_name(mat["path"], mat["path"])
    signode += addnodes.desc_sig_space()
    signode += addnodes.desc_annotation("", "(Earthly Artifact)")
    return sig


def setup(app: Sphinx):
    app.add_object_type(  # type: ignore
        "earthly-target",
        "earthly-target",
        indextemplate="pair: earthly target; %s",
        parse_node=annotator("Earthly target"),
    )
    app.add_object_type(  # type: ignore
        "script",
        "script",
        indextemplate="pair: shell script; %s",
        parse_node=annotator("shell script"),
    )
    app.add_object_type(  # type: ignore
        "earthly-artifact",
        "earthly-artifact",
        indextemplate="pair: earthly artifact; %s",
        parse_node=parse_earthly_artifact,
    )
    app.add_object_type(  # type: ignore
        "file",
        "file",
        indextemplate="repository file; %s",
        parse_node=annotator("repository file"),
    )



"""
Extremely basic subprocess control
"""

import argparse
import json
import os
import random
import signal
import subprocess
import sys
import time
import traceback
from datetime import datetime, timedelta
from pathlib import Path
from typing import TYPE_CHECKING, NoReturn, Sequence, Union, cast

if TYPE_CHECKING:
    from typing import (Literal, NamedTuple, TypedDict)

INTERUPT_SIGNAL = signal.SIGINT if os.name != 'nt' else signal.CTRL_C_SIGNAL


def create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser('proc-ctl')
    grp = parser.add_subparsers(title='Commands',
                                dest='command',
                                metavar='<subcommand>')

    start = grp.add_parser('start', help='Start a new subprocess')
    start.add_argument('--ctl-dir',
                       help='The control directory for the subprocess',
                       required=True,
                       type=Path)
    start.add_argument('--cwd',
                       help='The new subdirectory of the spawned process',
                       type=Path)
    start.add_argument(
        '--spawn-wait',
        help='Number of seconds to wait for child to be running',
        type=float,
        default=3)
    start.add_argument('child_command',
                       nargs='+',
                       help='The command to execute',
                       metavar='<command> [args...]')

    stop = grp.add_parser('stop', help='Stop a running subprocess')
    stop.add_argument('--ctl-dir',
                      help='The control directory for the subprocess',
                      required=True,
                      type=Path)
    stop.add_argument('--stop-wait',
                      help='Number of seconds to wait for stopping',
                      type=float,
                      default=5)
    stop.add_argument('--if-not-running',
                      help='Action to take if the child is not running',
                      choices=['fail', 'ignore'],
                      default='fail')

    ll_run = grp.add_parser('__run')
    ll_run.add_argument('--ctl-dir', type=Path, required=True)
    ll_run.add_argument('child_command', nargs='+')

    return parser


if TYPE_CHECKING:
    StartCommandArgs = NamedTuple('StartCommandArgs', [
        ('command', Literal['start']),
        ('ctl_dir', Path),
        ('cwd', Path),
        ('child_command', Sequence[str]),
        ('spawn_wait', float),
    ])

    StopCommandArgs = NamedTuple('StopCommandArgs', [
        ('command', Literal['stop']),
        ('ctl_dir', Path),
        ('stop_wait', float),
        ('if_not_running', Literal['fail', 'ignore']),
    ])

    _RunCommandArgs = NamedTuple('_RunCommandArgs', [
        ('command', Literal['__run']),
        ('child_command', Sequence[str]),
        ('ctl_dir', Path),
    ])

    CommandArgs = Union[StartCommandArgs, StopCommandArgs, _RunCommandArgs]

    _ResultType = TypedDict('_ResultType', {
        'exit': 'str | int | None',
        'error': 'str | None'
    })


def parse_argv(argv: 'Sequence[str]') -> 'CommandArgs':
    parser = create_parser()
    args = parser.parse_args(argv)
    return cast('CommandArgs', args)


class _ChildControl:

    def __init__(self, ctl_dir: Path) -> None:
        self._ctl_dir = ctl_dir

    @property
    def pid_file(self):
        """The file containing the child PID"""
        return self._ctl_dir / 'pid.txt'

    @property
    def result_file(self):
        """The file containing the exit result"""
        return self._ctl_dir / 'exit.json'

    def set_pid(self, pid: int):
        write_text(self.pid_file, str(pid))

    def get_pid(self) -> 'int | None':
        try:
            txt = self.pid_file.read_text()
        except FileNotFoundError:
            return None
        return int(txt)

    def set_exit(self, exit: 'str | int | None', error: 'str | None') -> None:
        write_text(self.result_file, json.dumps({
            'exit': exit,
            'error': error
        }))
        remove_file(self.pid_file)

    def get_result(self) -> 'None | _ResultType':
        try:
            txt = self.result_file.read_text()
        except FileNotFoundError:
            return None
        return json.loads(txt)

    def clear_result(self) -> None:
        remove_file(self.result_file)


def _start(args: 'StartCommandArgs') -> int:
    ll_run_cmd = [
        sys.executable,
        '-u',
        '--',
        __file__,
        '__run',
        '--ctl-dir={}'.format(args.ctl_dir),
        '--',
        *args.child_command,
    ]
    args.ctl_dir.mkdir(exist_ok=True, parents=True)
    child = _ChildControl(args.ctl_dir)
    if child.get_pid() is not None:
        raise RuntimeError('Child process is already running [PID {}]'.format(
            child.get_pid()))
    child.clear_result()
    # Spawn the child controller
    subprocess.Popen(
        ll_run_cmd,
        cwd=args.cwd,
        stderr=subprocess.STDOUT,
        stdout=args.ctl_dir.joinpath('runner-output.txt').open('wb'),
        stdin=subprocess.DEVNULL)
    expire = datetime.now() + timedelta(seconds=args.spawn_wait)
    # Wait for the PID to appear
    while child.get_pid() is None and child.get_result() is None:
        if expire < datetime.now():
            break
        time.sleep(0.1)
    # Check that it actually spawned
    if child.get_pid() is None:
        result = child.get_result()
        if result is None:
            raise RuntimeError('Failed to spawn child runner?')
        if result['error']:
            print(result['error'], file=sys.stderr)
        raise RuntimeError('Child exited immediately [Exited {}]'.format(
            result['exit']))
    # Wait to see that it is still running after --spawn-wait seconds
    while child.get_result() is None:
        if expire < datetime.now():
            break
        time.sleep(0.1)
    # A final check to see if it is running
    result = child.get_result()
    if result is not None:
        if result['error']:
            print(result['error'], file=sys.stderr)
        raise RuntimeError('Child exited prematurely [Exited {}]'.format(
            result['exit']))
    return 0


def _stop(args: 'StopCommandArgs') -> int:
    child = _ChildControl(args.ctl_dir)
    pid = child.get_pid()
    if pid is None:
        if args.if_not_running == 'fail':
            raise RuntimeError('Child process is not running')
        elif args.if_not_running == 'ignore':
            # Nothing to do
            return 0
        else:
            assert False
    os.kill(pid, INTERUPT_SIGNAL)
    expire_at = datetime.now() + timedelta(seconds=args.stop_wait)
    while expire_at > datetime.now() and child.get_result() is None:
        time.sleep(0.1)
    result = child.get_result()
    if result is None:
        raise RuntimeError(
            'Child process did not exit within the grace period')
    return 0


def __run(args: '_RunCommandArgs') -> int:
    this = _ChildControl(args.ctl_dir)
    try:
        pipe = subprocess.Popen(
            args.child_command,
            stdout=args.ctl_dir.joinpath('child-output.txt').open('wb'),
            stderr=subprocess.STDOUT,
            stdin=subprocess.DEVNULL)
    except:
        this.set_exit('spawn-failed', traceback.format_exc())
        raise
    this.set_pid(pipe.pid)
    retc = None
    try:
        while 1:
            try:
                retc = pipe.wait(0.5)
            except subprocess.TimeoutExpired:
                pass
            except KeyboardInterrupt:
                pipe.send_signal(INTERUPT_SIGNAL)
            if retc is not None:
                break
    finally:
        this.set_exit(retc, None)
    return 0


def write_text(fpath: Path, content: str):
    """
    "Atomically" write a new file.

    This writes the given ``content`` into a temporary file, then renames that
    file into place. This prevents readers from seeing a partial read.
    """
    tmp = fpath.with_name(fpath.name + '.tmp')
    remove_file(tmp)
    tmp.write_text(content)
    os.sync()
    remove_file(fpath)
    tmp.rename(fpath)


def remove_file(fpath: Path):
    """
    Safely remove a file.

    Because Win32, deletes are asynchronous, so we rename to a random filename,
    then delete that file. This ensures the file is "out of the way", even if
    it takes some time to delete.
    """
    delname = fpath.with_name(fpath.name + '.delete-' +
                              str(random.randint(0, 999999)))
    try:
        fpath.rename(delname)
    except FileNotFoundError:
        return
    delname.unlink()


def main(argv: 'Sequence[str]') -> int:
    args = parse_argv(argv)
    if args.command == 'start':
        return _start(args)
    if args.command == '__run':
        return __run(args)
    if args.command == 'stop':
        return _stop(args)
    return 0


def start_main() -> NoReturn:
    sys.exit(main(sys.argv[1:]))


if __name__ == '__main__':
    start_main()



#!/usr/bin/env python
#
# Copyright 2009-present MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generate test functions for use with mock_server_t.

Defines functions like future_cursor_next in future-functions.h and
future-functions.c, which defer a libmongoc operation to a background thread
via functions like background_cursor_next. Also defines functions like
future_value_set_bson_ptr and future_value_get_bson_ptr which support the
future / background functions, and functions like future_get_bson_ptr which
wait for a future to resolve, then return its value.

These future functions are used in conjunction with mock_server_t to
conveniently test libmongoc wire protocol operations.

Written for Python 2.6+, requires Jinja 2 for templating.
"""

from collections import namedtuple
from os.path import basename, dirname, join as joinpath, normpath

# Please "pip install jinja2".
from jinja2 import Environment, FileSystemLoader

this_dir = dirname(__file__)
template_dir = joinpath(this_dir, 'future_function_templates')
mock_server_dir = normpath(
    joinpath(this_dir, '../src/libmongoc/tests/mock_server'))

# Add additional types here. Use typedefs for derived types so they can
# be named with one symbol.
typedef = namedtuple("typedef", ["name", "typedef"])

# These are typedef'ed if necessary in future-value.h, and added to the union
# of possible future_value_t.value types. future_value_t getters and setters
# are generated for all types, as well as future_t getters.
typedef_list = [
    # Fundamental.
    typedef("bool", None),
    typedef("char_ptr", "char *"),
    typedef("char_ptr_ptr", "char **"),
    typedef("int", None),
    typedef("int64_t", None),
    typedef("size_t", None),
    typedef("ssize_t", None),
    typedef("uint32_t", None),
    typedef("void_ptr",  "void *"),

    # Const fundamental.
    typedef("const_char_ptr", "const char *"),
    typedef("bool_ptr", "bool *"),

    # libbson.
    typedef("bson_error_ptr", "bson_error_t *"),
    typedef("bson_ptr", "bson_t *"),

    # Const libbson.
    typedef("const_bson_ptr", "const bson_t *"),
    typedef("const_bson_ptr_ptr", "const bson_t **"),

    # libmongoc.
    typedef("mongoc_async_ptr", "mongoc_async_t *"),
    typedef("mongoc_bulk_operation_ptr", "mongoc_bulk_operation_t *"),
    typedef("mongoc_client_ptr", "mongoc_client_t *"),
    typedef("mongoc_client_pool_ptr", "mongoc_client_pool_t *"),
    typedef("mongoc_collection_ptr", "mongoc_collection_t *"),
    typedef("mongoc_cluster_ptr", "mongoc_cluster_t *"),
    typedef("mongoc_cmd_parts_ptr", "mongoc_cmd_parts_t *"),
    typedef("mongoc_cursor_ptr", "mongoc_cursor_t *"),
    typedef("mongoc_database_ptr", "mongoc_database_t *"),
    typedef("mongoc_gridfs_file_ptr", "mongoc_gridfs_file_t *"),
    typedef("mongoc_gridfs_ptr", "mongoc_gridfs_t *"),
    typedef("mongoc_insert_flags_t", None),
    typedef("mongoc_iovec_ptr", "mongoc_iovec_t *"),
    typedef("mongoc_server_stream_ptr", "mongoc_server_stream_t *"),
    typedef("mongoc_query_flags_t", None),
    typedef("const_mongoc_index_opt_t", "const mongoc_index_opt_t *"),
    typedef("mongoc_server_description_ptr", "mongoc_server_description_t *"),
    typedef("mongoc_ss_optype_t", None),
    typedef("mongoc_topology_ptr", "mongoc_topology_t *"),
    typedef("mongoc_write_concern_ptr", "mongoc_write_concern_t *"),
    typedef("mongoc_change_stream_ptr", "mongoc_change_stream_t *"),
    typedef("mongoc_remove_flags_t", None),

    # Const libmongoc.
    typedef("const_mongoc_find_and_modify_opts_ptr",
            "const mongoc_find_and_modify_opts_t *"),
    typedef("const_mongoc_iovec_ptr", "const mongoc_iovec_t *"),
    typedef("const_mongoc_read_prefs_ptr", "const mongoc_read_prefs_t *"),
    typedef("const_mongoc_write_concern_ptr",
            "const mongoc_write_concern_t *"),
    typedef("const_mongoc_ss_log_context_ptr",
            "const mongoc_ss_log_context_t *"),
]

type_list = [T.name for T in typedef_list]
type_list_with_void = type_list + ['void']

param = namedtuple("param", ["type_name", "name"])
future_function = namedtuple("future_function", ["ret_type", "name", "params"])

# Add additional functions to be tested here. For a name like "cursor_next", we
# generate two functions: future_cursor_next to prepare the future_t and launch
# a background thread, and background_cursor_next to run on the thread and
# resolve the future.
future_functions = [
    future_function("void",
                    "mongoc_async_run",
                    [param("mongoc_async_ptr", "async")]),

    future_function("uint32_t",
                    "mongoc_bulk_operation_execute",
                    [param("mongoc_bulk_operation_ptr", "bulk"),
                     param("bson_ptr", "reply"),
                     param("bson_error_ptr", "error")]),

    future_function("bool",
                    "mongoc_database_read_command_with_opts",
                    [param("mongoc_database_ptr", "database"),
                     param("const_bson_ptr", "command"),
                     param("const_mongoc_read_prefs_ptr", "read_prefs"),
                     param("const_bson_ptr", "opts"),
                     param("bson_ptr", "reply"),
                     param("bson_error_ptr", "error")]),

    future_function("bool",
                    "mongoc_database_read_write_command_with_opts",
                    [param("mongoc_database_ptr", "database"),
                     param("const_bson_ptr", "command"),
                     param("const_mongoc_read_prefs_ptr", "read_prefs"),
                     param("const_bson_ptr", "opts"),
                     param("bson_ptr", "reply"),
                     param("bson_error_ptr", "error")]),

    future_function("bool",
                    "mongoc_database_write_command_with_opts",
                    [param("mongoc_database_ptr", "database"),
                     param("const_bson_ptr", "command"),
                     param("const_bson_ptr", "opts"),
                     param("bson_ptr", "reply"),
                     param("bson_error_ptr", "error")]),

    future_function("bool",
                    "mongoc_client_command_simple",
                    [param("mongoc_client_ptr", "client"),
                     param("const_char_ptr", "db_name"),
                     param("const_bson_ptr", "command"),
                     param("const_mongoc_read_prefs_ptr", "read_prefs"),
                     param("bson_ptr", "reply"),
                     param("bson_error_ptr", "error")]),

    future_function("bool",
                    "mongoc_client_command_with_opts",
                    [param("mongoc_client_ptr", "client"),
                     param("const_char_ptr", "db_name"),
                     param("const_bson_ptr", "command"),
                     param("const_mongoc_read_prefs_ptr", "read_prefs"),
                     param("const_bson_ptr", "opts"),
                     param("bson_ptr", "reply"),
                     param("bson_error_ptr", "error")]),

    future_function("bool",
                    "mongoc_client_read_command_with_opts",
                    [param("mongoc_client_ptr", "client"),
                     param("const_char_ptr", "db_name"),
                     param("const_bson_ptr", "command"),
                     param("const_mongoc_read_prefs_ptr", "read_prefs"),
                     param("const_bson_ptr", "opts"),
                     param("bson_ptr", "reply"),
                     param("bson_error_ptr", "error")]),

    future_function("bool",
                    "mongoc_client_write_command_with_opts",
                    [param("mongoc_client_ptr", "client"),
                     param("const_char_ptr", "db_name"),
                     param("const_bson_ptr", "command"),
                     param("const_bson_ptr", "opts"),
                     param("bson_ptr", "reply"),
                     param("bson_error_ptr", "error")]),

    future_function("bool",
                    "mongoc_client_read_write_command_with_opts",
                    [param("mongoc_client_ptr", "client"),
                     param("const_char_ptr", "db_name"),
                     param("const_bson_ptr", "command"),
                     param("const_mongoc_read_prefs_ptr", "read_prefs"),
                     param("const_bson_ptr", "opts"),
                     param("bson_ptr", "reply"),
                     param("bson_error_ptr", "error")]),

    future_function("void",
                    "mongoc_client_kill_cursor",
                    [param("mongoc_client_ptr", "client"),
                     param("int64_t", "cursor_id")]),

    future_function("mongoc_change_stream_ptr",
                    "mongoc_client_watch",
                    [param("mongoc_client_ptr", "client"),
                     param("const_bson_ptr", "pipeline"),
                     param("const_bson_ptr", "opts")]),

    future_function("mongoc_cursor_ptr",
                    "mongoc_collection_aggregate",
                    [param("mongoc_collection_ptr", "collection"),
                     param("mongoc_query_flags_t", "flags"),
                     param("const_bson_ptr", "pipeline"),
                     param("const_bson_ptr", "options"),
                     param("const_mongoc_read_prefs_ptr", "read_prefs")]),

    future_function("int64_t",
                    "mongoc_collection_count",
                    [param("mongoc_collection_ptr", "collection"),
                     param("mongoc_query_flags_t", "flags"),
                     param("const_bson_ptr", "query"),
                     param("int64_t", "skip"),
                     param("int64_t", "limit"),
                     param("const_mongoc_read_prefs_ptr", "read_prefs"),
                     param("bson_error_ptr", "error")]),

    future_function("int64_t",
                    "mongoc_collection_count_with_opts",
                    [param("mongoc_collection_ptr", "collection"),
                     param("mongoc_query_flags_t", "flags"),
                     param("const_bson_ptr", "query"),
                     param("int64_t", "skip"),
                     param("int64_t", "limit"),
                     param("const_bson_ptr", "opts"),
                     param("const_mongoc_read_prefs_ptr", "read_prefs"),
                     param("bson_error_ptr", "error")]),

    future_function("bool",
                    "mongoc_collection_create_index_with_opts",
                    [param("mongoc_collection_ptr", "collection"),
                     param("const_bson_ptr", "keys"),
                     param("const_mongoc_index_opt_t", "opt"),
                     param("const_bson_ptr", "opts"),
                     param("bson_ptr", "reply"),
                     param("bson_error_ptr", "error")]),

    future_function("bool",
                    "mongoc_collection_drop_index_with_opts",
                    [param("mongoc_collection_ptr", "collection"),
                     param("const_char_ptr", "index_name"),
                     param("const_bson_ptr", "opts"),
                     param("bson_error_ptr", "error")]),

    future_function("bool",
                    "mongoc_collection_drop_with_opts",
                    [param("mongoc_collection_ptr", "collection"),
                     param("const_bson_ptr", "opts"),
                     param("bson_error_ptr", "error")]),

    future_function("bool",
                    "mongoc_collection_find_and_modify_with_opts",
                    [param("mongoc_collection_ptr", "collection"),
                     param("const_bson_ptr", "query"),
                     param("const_mongoc_find_and_modify_opts_ptr", "opts"),
                     param("bson_ptr", "reply"),
                     param("bson_error_ptr", "error")]),

    future_function("bool",
                    "mongoc_collection_find_and_modify",
                    [param("mongoc_collection_ptr", "collection"),
                     param("const_bson_ptr", "query"),
                     param("const_bson_ptr", "sort"),
                     param("const_bson_ptr", "update"),
                     param("const_bson_ptr", "fields"),
                     param("bool", "_remove"),
                     param("bool", "upsert"),
                     param("bool", "_new"),
                     param("bson_ptr", "reply"),
                     param("bson_error_ptr", "error")]),

    future_function("mongoc_cursor_ptr",
                    "mongoc_collection_find_indexes_with_opts",
                    [param("mongoc_collection_ptr", "collection"),
                     param("const_bson_ptr", "opts")]),

    future_function("bool",
                    "mongoc_collection_stats",
                    [param("mongoc_collection_ptr", "collection"),
                     param("const_bson_ptr", "options"),
                     param("bson_ptr", "stats"),
                     param("bson_error_ptr", "error")]),

    future_function("bool",
                    "mongoc_collection_insert_many",
                    [param("mongoc_collection_ptr", "collection"),
                     param("const_bson_ptr_ptr", "documents"),
                     param("size_t", "n_documents"),
                     param("const_bson_ptr", "opts"),
                     param("bson_ptr", "reply"),
                     param("bson_error_ptr", "error")]),

    future_function("bool",
                    "mongoc_collection_insert_one",
                    [param("mongoc_collection_ptr", "collection"),
                     param("const_bson_ptr", "document"),
                     param("const_bson_ptr", "opts"),
                     param("bson_ptr", "reply"),
                     param("bson_error_ptr", "error")]),

    future_function("bool",
                    "mongoc_collection_read_command_with_opts",
                    [param("mongoc_collection_ptr", "collection"),
                     param("const_bson_ptr", "command"),
                     param("const_mongoc_read_prefs_ptr", "read_prefs"),
                     param("const_bson_ptr", "opts"),
                     param("bson_ptr", "reply"),
                     param("bson_error_ptr", "error")]),

    future_function("bool",
                    "mongoc_collection_read_write_command_with_opts",
                    [param("mongoc_collection_ptr", "collection"),
                     param("const_bson_ptr", "command"),
                     param("const_mongoc_read_prefs_ptr", "read_prefs"),
                     param("const_bson_ptr", "opts"),
                     param("bson_ptr", "reply"),
                     param("bson_error_ptr", "error")]),

    future_function("bool",
                    "mongoc_collection_write_command_with_opts",
                    [param("mongoc_collection_ptr", "collection"),
                     param("const_bson_ptr", "command"),
                     param("const_bson_ptr", "opts"),
                     param("bson_ptr", "reply"),
                     param("bson_error_ptr", "error")]),

    future_function("bool",
                    "mongoc_collection_insert_bulk",
                    [param("mongoc_collection_ptr", "collection"),
                     param("mongoc_insert_flags_t", "flags"),
                     param("const_bson_ptr_ptr", "documents"),
                     param("uint32_t", "n_documents"),
                     param("const_mongoc_write_concern_ptr", "write_concern"),
                     param("bson_error_ptr", "error")]),

    future_function("bool",
                    "mongoc_cluster_run_command_parts",
                    [param("mongoc_cluster_ptr", "cluster"),
                     param("mongoc_server_stream_ptr", "server_stream"),
                     param("mongoc_cmd_parts_ptr", "parts"),
                     param("bson_ptr", "reply"),
                     param("bson_error_ptr", "error")]),

    future_function("void",
                    "mongoc_cursor_destroy",
                    [param("mongoc_cursor_ptr", "cursor")]),

    future_function("bool",
                    "mongoc_cursor_next",
                    [param("mongoc_cursor_ptr", "cursor"),
                     param("const_bson_ptr_ptr", "doc")]),

    future_function("char_ptr_ptr",
                    "mongoc_client_get_database_names_with_opts",
                    [param("mongoc_client_ptr", "client"),
                     param("const_bson_ptr", "opts"),
                     param("bson_error_ptr", "error")]),

    future_function("mongoc_server_description_ptr",
                    "mongoc_client_select_server",
                    [param("mongoc_client_ptr", "client"),
                     param("bool", "for_writes"),
                     param("const_mongoc_read_prefs_ptr", "prefs"),
                     param("bson_error_ptr", "error")]),

    future_function("void",
                    "mongoc_client_destroy",
                    [param("mongoc_client_ptr", "client")]),

    future_function("void",
                    "mongoc_client_pool_destroy",
                    [param("mongoc_client_pool_ptr", "pool")]),

    future_function("bool",
                    "mongoc_database_command_simple",
                    [param("mongoc_database_ptr", "database"),
                     param("bson_ptr", "command"),
                     param("const_mongoc_read_prefs_ptr", "read_prefs"),
                     param("bson_ptr", "reply"),
                     param("bson_error_ptr", "error")]),

    future_function("bool",
                    "mongoc_database_drop_with_opts",
                    [param("mongoc_database_ptr", "database"),
                     param("const_bson_ptr", "opts"),
                     param("bson_error_ptr", "error")]),

    future_function("char_ptr_ptr",
                    "mongoc_database_get_collection_names_with_opts",
                    [param("mongoc_database_ptr", "database"),
                     param("const_bson_ptr", "opts"),
                     param("bson_error_ptr", "error")]),

    future_function("mongoc_change_stream_ptr",
                    "mongoc_database_watch",
                    [param("mongoc_database_ptr", "database"),
                     param("const_bson_ptr", "pipeline"),
                     param("const_bson_ptr", "opts")]),

    future_function("ssize_t",
                    "mongoc_gridfs_file_readv",
                    [param("mongoc_gridfs_file_ptr", "file"),
                     param("mongoc_iovec_ptr", "iov"),
                     param("size_t", "iovcnt"),
                     param("size_t", "min_bytes"),
                     param("uint32_t", "timeout_msec")]),

    future_function("mongoc_gridfs_file_ptr",
                    "mongoc_gridfs_find_one",
                    [param("mongoc_gridfs_ptr", "gridfs"),
                     param("const_bson_ptr", "query"),
                     param("bson_error_ptr", "error")]),

    future_function("bool",
                    "mongoc_gridfs_file_remove",
                    [param("mongoc_gridfs_file_ptr", "file"),
                     param("bson_error_ptr", "error")]),

    future_function("int",
                    "mongoc_gridfs_file_seek",
                    [param("mongoc_gridfs_file_ptr", "file"),
                     param("int64_t", "delta"),
                     param("int", "whence")]),

    future_function("ssize_t",
                    "mongoc_gridfs_file_writev",
                    [param("mongoc_gridfs_file_ptr", "file"),
                     param("const_mongoc_iovec_ptr", "iov"),
                     param("size_t", "iovcnt"),
                     param("uint32_t", "timeout_msec")]),

    future_function("mongoc_gridfs_file_ptr",
                    "mongoc_gridfs_find_one_with_opts",
                    [param("mongoc_gridfs_ptr", "gridfs"),
                     param("const_bson_ptr", "filter"),
                     param("const_bson_ptr", "opts"),
                     param("bson_error_ptr", "error")]),

    future_function("mongoc_server_description_ptr",
                    "mongoc_topology_select",
                    [param("mongoc_topology_ptr", "topology"),
                     param("mongoc_ss_optype_t", "optype"),
                     param("const_mongoc_ss_log_context_ptr", "log_context"),
                     param("const_mongoc_read_prefs_ptr", "read_prefs"),
                     param("bool_ptr", "must_use_primary"),
                     param("bson_error_ptr", "error")]),

    future_function("mongoc_gridfs_ptr",
                    "mongoc_client_get_gridfs",
                    [param("mongoc_client_ptr", "client"),
                     param("const_char_ptr", "db"),
                     param("const_char_ptr", "prefix"),
                     param("bson_error_ptr", "error")]),

    future_function("mongoc_change_stream_ptr",
                    "mongoc_collection_watch",
                    [param("mongoc_collection_ptr", "coll"),
                     param("const_bson_ptr", "pipeline"),
                     param("const_bson_ptr", "opts")]),

    future_function("bool",
                    "mongoc_change_stream_next",
                    [param("mongoc_change_stream_ptr", "stream"),
                     param("const_bson_ptr_ptr", "bson")]),

    future_function("void",
                    "mongoc_change_stream_destroy",
                    [param("mongoc_change_stream_ptr", "stream")]),

    future_function("bool",
                    "mongoc_collection_delete_one",
                    [param("mongoc_collection_ptr", "coll"),
                     param("const_bson_ptr", "selector"),
                     param("const_bson_ptr", "opts"),
                     param("bson_ptr", "reply"),
                     param("bson_error_ptr", "error")]),

    future_function("bool",
                    "mongoc_collection_delete_many",
                    [param("mongoc_collection_ptr", "coll"),
                     param("const_bson_ptr", "selector"),
                     param("const_bson_ptr", "opts"),
                     param("bson_ptr", "reply"),
                     param("bson_error_ptr", "error")]),

    future_function("bool",
                    "mongoc_collection_remove",
                    [param("mongoc_collection_ptr", "coll"),
                     param("mongoc_remove_flags_t", "flags"),
                     param("const_bson_ptr", "selector"),
                     param("const_mongoc_write_concern_ptr", "write_concern"),
                     param("bson_error_ptr", "error")]),

    future_function("bool",
                    "mongoc_collection_update_one",
                    [param("mongoc_collection_ptr", "coll"),
                     param("const_bson_ptr", "selector"),
                     param("const_bson_ptr", "update"),
                     param("const_bson_ptr", "opts"),
                     param("bson_ptr", "reply"),
                     param("bson_error_ptr", "error")]),

    future_function("bool",
                    "mongoc_collection_update_many",
                    [param("mongoc_collection_ptr", "coll"),
                     param("const_bson_ptr", "selector"),
                     param("const_bson_ptr", "update"),
                     param("const_bson_ptr", "opts"),
                     param("bson_ptr", "reply"),
                     param("bson_error_ptr", "error")]),

    future_function("bool",
                    "mongoc_collection_replace_one",
                    [param("mongoc_collection_ptr", "coll"),
                     param("const_bson_ptr", "selector"),
                     param("const_bson_ptr", "replacement"),
                     param("const_bson_ptr", "opts"),
                     param("bson_ptr", "reply"),
                     param("bson_error_ptr", "error")]),

    future_function("int64_t",
                    "mongoc_collection_count_documents",
                    [param("mongoc_collection_ptr", "coll"),
                     param("const_bson_ptr", "filter"),
                     param("const_bson_ptr", "opts"),
                     param("const_mongoc_read_prefs_ptr", "read_prefs"),
                     param("bson_ptr", "reply"),
                     param("bson_error_ptr", "error")]),

    future_function("int64_t",
                    "mongoc_collection_estimated_document_count",
                    [param("mongoc_collection_ptr", "coll"),
                     param("const_bson_ptr", "opts"),
                     param("const_mongoc_read_prefs_ptr", "read_prefs"),
                     param("bson_ptr", "reply"),
                     param("bson_error_ptr", "error")]),
]

for fn in future_functions:
    if fn.ret_type not in type_list_with_void:
        raise Exception('bad type "%s"\n\nin %s' % (fn.ret_type, fn))
    for p in fn.params:
        if p.type_name not in type_list:
            raise Exception('bad type "%s"\n\nin %s' % (p.type_name, fn))

header_comment = """/**************************************************
 *
 * Generated by build/%s.
 *
 * DO NOT EDIT THIS FILE.
 *
 *************************************************/
/* clang-format off */""" % basename(__file__)


def future_function_name(fn):
    if fn.name.startswith('mongoc'):
        # E.g. future_cursor_next().
        return 'future' + fn.name[len('mongoc'):]
    else:
        # E.g. future__mongoc_client_kill_cursor().
        return 'future_' + fn.name


env = Environment(loader=FileSystemLoader(template_dir))
env.filters['future_function_name'] = future_function_name

files = ["future.h",
         "future.c",
         "future-value.h",
         "future-value.c",
         "future-functions.h",
         "future-functions.c"]

for file_name in files:
    print(file_name)
    with open(joinpath(mock_server_dir, file_name), 'w+') as f:
        t = env.get_template(file_name + ".template")
        f.write(t.render(globals()))
        f.write('\n')



from shrub.v3.evg_command import archive_targz_pack
from shrub.v3.evg_command import s3_put

from config_generator.etc.function import Function
from config_generator.etc.utils import bash_exec


class UploadMOArtifacts(Function):
    name = 'upload-mo-artifacts'
    commands = [
        bash_exec(
            working_dir='mongoc',
            script='''\
                set -o errexit
                declare dir="MO"
                if [[ -d "/cygdrive/c/data/mo" ]]; then
                    dir="/cygdrive/c/data/mo"
                fi
                if [[ -d "$dir" ]]; then
                    find "$dir" -name \\*.log | xargs tar czf mongodb-logs.tar.gz
                fi
            '''
        ),
        s3_put(
            aws_key='${aws_key}',
            aws_secret='${aws_secret}',
            bucket='mciuploads',
            content_type='${content_type|application/x-gzip}',
            display_name='mongodb-logs.tar.gz',
            local_file='mongoc/mongodb-logs.tar.gz',
            optional=True,
            permissions='public-read',
            remote_file='mongo-c-driver/${build_variant}/${revision}/${version_id}/${build_id}/logs/${task_id}-${execution}-mongodb-logs.tar.gz',
        ),
        s3_put(
            aws_key='${aws_key}',
            aws_secret='${aws_secret}',
            bucket='mciuploads',
            content_type='${content_type|text/plain}',
            display_name='orchestration.log',
            local_file='mongoc/MO/server.log',
            optional=True,
            permissions='public-read',
            remote_file='mongo-c-driver/${build_variant}/${revision}/${version_id}/${build_id}/logs/${task_id}-${execution}-orchestration.log',
        ),
        bash_exec(
            working_dir='mongoc',
            script='''\
                set -o errexit
                # Find all core files from mongodb in orchestration and move to mongoc
                declare dir="MO"
                if [[ -d "/cygdrive/c/data/mo" ]]; then
                    dir="/cygdrive/c/data/mo"
                fi
                declare mdmp_dir="$dir"
                if [[ -d "/cygdrive/c/mongodb" ]]; then
                    mdmp_dir="/cygdrive/c/mongodb"
                fi
                for core_file in $(find -H "$dir" "$mdmp_dir" \\( -name "*.core" -o -name "*.mdmp" \\) 2> /dev/null); do
                    declare base_name
                    base_name="$(echo "$core_file" | sed "s/.*\\///")"
                    # Move file if it does not already exist
                    if [[ ! -f "$base_name" ]]; then
                        mv "$core_file" .
                    fi
                done
            '''
        ),
        archive_targz_pack(
            target='mongo-coredumps.tgz',
            source_dir='mongoc',
            include=[
                './**.core',
                './**.mdmp',
            ],
        ),
        s3_put(
            aws_key='${aws_key}',
            aws_secret='${aws_secret}',
            bucket='mciuploads',
            content_type='${content_type|application/x-gzip}',
            display_name='Core Dumps - Execution ${execution}',
            local_file='mongo-coredumps.tgz',
            optional=True,
            permissions='public-read',
            remote_file='mongo-c-driver/${build_variant}/${revision}/${version_id}/${build_id}/coredumps/${task_id}-${execution}-coredumps.log',
        ),
    ]

    @classmethod
    def call(cls, **kwargs):
        return cls.default_call(**kwargs)


def functions():
    return UploadMOArtifacts.defn()



# -*- coding: utf-8 -*-
import http.client
import os.path
import sys
import urllib.request
from pathlib import Path
from typing import Any, Dict

try:
    from sphinx.builders.dirhtml import DirectoryHTMLBuilder
except ImportError:
    # Try importing from older Sphinx version path.
    from sphinx.builders.html import DirectoryHTMLBuilder

from docutils.parsers.rst import directives, Directive
from sphinx.application import Sphinx
from sphinx.application import logger as sphinx_log
from sphinx.config import Config

has_sphinx_design = False
try:
    # Try to import sphinx-design to include directives for HTML pages (e.g. tabs and dropdowns).
    # sphinx-design is not required for building man pages.
    # python-sphinx-design is not currently available on EPEL. The package for EPEL includes man pages.
    from sphinx_design.dropdown import DropdownDirective
    has_sphinx_design = True
except ImportError:
    pass

# Ensure we can import "mongoc" extension module.
this_path = os.path.dirname(__file__)
sys.path.append(this_path)
sys.path.append(os.path.normpath(os.path.join(this_path, "../../../build/sphinx")))

from mongoc_common import *

extensions = [
    "mongoc",
    "sphinx.ext.intersphinx",
    # NOTE: We use our own "minimal" CMake domain that lets us refer to external
    # objects from the CMake inventory, but provides no other features. The
    # build *could* otherwise use sphinxcontrib-moderncmakedomain, which is
    # more full-featured, but it is not (currently) available in repositories for
    # package building.
    # "sphinxcontrib.moderncmakedomain",
    "cmakerefdomain",
    "sphinx.ext.mathjax",
]

if has_sphinx_design:
    extensions.append("sphinx_design")

# General information about the project.
project = "libmongoc"
copyright = "2009-present, MongoDB, Inc."
author = "MongoDB, Inc"

version_path = os.path.join(os.path.dirname(__file__), "../../..", "VERSION_CURRENT")
version = open(version_path).read().strip()

# The extension requires the "base" to contain '%s' exactly once, but we never intend to use it though

language = "en"
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
master_doc = "index"
html_static_path = ["static"]

# Set an empty list of disabled reftypes.
# Sphinx 5.0 disables "std:doc" by default.
# Many documentation references use :doc:
intersphinx_disabled_reftypes = []
# Give the peer 30sec to give us the inventory:
intersphinx_timeout = 30

intersphinx_mapping = {
    "sphinx": ("https://www.sphinx-doc.org/en/master", "includes/sphinx.inv"),
    "python": ("https://docs.python.org/3", "includes/python.inv"),
    "bson": ("https://www.mongoc.org/libbson/current", "includes/libbson.inv"),
    "cmake": ("https://cmake.org/cmake/help/latest", "includes/cmake.inv"),
}

_UPDATE_KEY = "update_external_inventories"


def _maybe_update_inventories(app: Sphinx):
    """
    We save Sphinx inventories for external projects saved within our own project
    so that we can support fully-offline builds. This is a convenience function
    to update those inventories automatically.

    This function will only have an effect if the appropriate command-line config
    value is defined.
    """
    prefix = "[libmongoc/doc/conf.py]"
    config = app.config
    if not config[_UPDATE_KEY]:
        sphinx_log.info(
            "%s Using existing intersphinx inventories. Refresh by running with ‘-D %s=1’",
            prefix,
            _UPDATE_KEY,
        )
        return
    for name, tup in intersphinx_mapping.items():
        urlbase, filename = tup
        url = f"{urlbase}/objects.inv"
        sphinx_log.info("%s Downloading external inventory for %s from [%s]", prefix, name, url)
        with urllib.request.urlopen(url) as req:
            req: http.client.HTTPResponse = req
            dest = Path(app.srcdir) / filename
            sphinx_log.info("%s Saving inventory [%s] to file [%s]", prefix, url, dest)
            with dest.open("wb") as out:
                buf = req.read(1024 * 4)
                while buf:
                    out.write(buf)
                    buf = req.read(1024 * 4)
        sphinx_log.info(
            "%s Inventory file [%s] was updated. Commit the result to save it for subsequent builds.",
            prefix,
            dest,
        )


# -- Options for HTML output ----------------------------------------------

html_theme = "furo"
html_title = html_shorttitle = "libmongoc %s" % version
# html_favicon = None
html_use_index = True

rst_prolog = rf"""
.. |qenc:is-experimental| replace::

    is part of the experimental
    `Queryable Encryption <queryable-encryption_>`_ API and may be subject
    to breaking changes in future releases.

.. |qenc:opt-is-experimental| replace::

    This option |qenc:is-experimental|

.. |qenc:api-is-experimental| replace::

    This API |qenc:is-experimental|

.. |qenc:range-is-experimental| replace::

    Range algorithm is experimental only and not intended for public use. It is subject to breaking changes.

.. _the findAndModify command:
    https://www.mongodb.com/docs/manual/reference/command/findAndModify/

.. |version| replace:: {version}
.. |version.pre| replace:: ``{version}``
.. |vversion| replace:: ``v{version}``

.. role:: bash(code)
    :language: bash

.. role:: batch(code)
    :language: batch

.. role:: c(code)
    :language: c

.. role:: cpp(code)
    :language: c++

.. role:: bolded-name(literal)
    :class: bolded-name

.. |libbson| replace:: :bolded-name:`libbson`
.. |libmongoc| replace:: :bolded-name:`libmongoc`
.. |mongo-c-driver| replace:: :bolded-name:`mongo-c-driver`

.. The CMake inventory mangles the names of its custom domain objects, for some reason?
   Offer these substitutions for simpler variable references:

.. |cmvar:CMAKE_BUILD_TYPE| replace::
    :cmake:variable:`CMAKE_BUILD_TYPE <variable:CMAKE_BUILD_TYPE>`

.. |cmvar:CMAKE_INSTALL_PREFIX| replace::
    :cmake:variable:`CMAKE_INSTALL_PREFIX <variable:CMAKE_INSTALL_PREFIX>`

.. |cmvar:CMAKE_PREFIX_PATH| replace::
    :cmake:variable:`CMAKE_PREFIX_PATH <variable:CMAKE_PREFIX_PATH>`

.. |cmcmd:find_package| replace::
    :cmake:command:`find_package() <command:find_package>`

.. |bson_t-storage-ptr| replace::
    non-``NULL`` pointer to `overwritable storage <bson_lifetimes_>`_ for a :symbol:`bson_t`

.. |bson_t-opt-storage-ptr| replace::
    maybe-``NULL`` pointer to `overwritable storage <bson_lifetimes_>`_ for a :symbol:`bson_t`

.. _mongodb_docs_cdriver: https://www.mongodb.com/docs/languages/c/c-driver/current/

.. _queryable-encryption: https://www.mongodb.com/docs/languages/c/c-driver/current/libmongoc/guides/queryable-encryption/

.. _authentication: https://www.mongodb.com/docs/languages/c/c-driver/current/libmongoc/authentication/

.. _bulk: https://www.mongodb.com/docs/languages/c/c-driver/current/libmongoc/guides/bulk/

.. _client-side-field-level-encryption: https://www.mongodb.com/docs/languages/c/c-driver/current/libmongoc/guides/client-side-field-level-encryption/

.. _configuring_tls: https://www.mongodb.com/docs/languages/c/c-driver/current/libmongoc/guides/configuring_tls/

.. _connection-pooling: https://www.mongodb.com/docs/languages/c/c-driver/current/libmongoc/guides/connection-pooling/

.. _in-use-encryption: https://www.mongodb.com/docs/languages/c/c-driver/current/libmongoc/guides/in-use-encryption/

.. _manage-collection-indexes: https://www.mongodb.com/docs/languages/c/c-driver/current/libmongoc/guides/manage-collection-indexes/

.. _setting_collation_order: https://www.mongodb.com/docs/languages/c/c-driver/current/libmongoc/guides/bulk/#setting-collation-order

.. _tutorial_crud_operations: https://www.mongodb.com/docs/languages/c/c-driver/current/libmongoc/tutorial/#basic-crud-operations

.. _bulk_operation_bypassing_document_validation: https://www.mongodb.com/docs/languages/c/c-driver/current/libmongoc/guides/bulk/#bulk-operation-bypassing-document-validation

.. _cursors_tailable: https://www.mongodb.com/docs/languages/c/c-driver/current/libmongoc/guides/cursors/#tailable-cursors

.. _bson_errors: https://www.mongodb.com/docs/languages/c/c-driver/current/libbson/tutorials/errors/

.. _bson_lifetimes: https://www.mongodb.com/docs/languages/c/c-driver/current/libbson/guides/lifetimes/

.. |ssl:naming| replace::
    Though some API names include the term "ssl", the C driver only support TLS protocols, which supersede SSL.

"""


def add_canonical_link(app: Sphinx, pagename: str, templatename: str, context: Dict[str, Any], doctree: Any):
    link = f'<link rel="canonical" href="https://www.mongoc.org/libmongoc/current/{pagename}/"/>'

    context["metatags"] = context.get("metatags", "") + link


if has_sphinx_design:
    class AdDropdown(DropdownDirective):
        """A sphinx-design dropdown that can also be an admonition."""

        option_spec = DropdownDirective.option_spec | {"admonition": directives.unchanged_required}

        def run(self):
            adm = self.options.get("admonition")
            if adm is not None:
                self.options.setdefault("class-container", []).extend(("admonition", adm))
                self.options.setdefault("class-title", []).append(f"admonition-title")
            return super().run()
else:
    class EmptyDirective(Directive):
        has_content = True
        def run(self):
            return []
        
has_add_css_file = True
        
def check_html_builder_requirements (app):
    if isinstance(app.builder, DirectoryHTMLBuilder):
        if not has_sphinx_design:
            raise RuntimeError("The sphinx-design package is required to build HTML documentation but was not detected. Install sphinx-design.")
        if not has_add_css_file:
            raise RuntimeError("A newer version of Sphinx is required to build HTML documentation with CSS files. Upgrade Sphinx to v3.5.0 or newer")

def setup(app: Sphinx):
    mongoc_common_setup(app)
    app.connect("builder-inited", check_html_builder_requirements)
    if has_sphinx_design:
        app.add_directive("ad-dropdown", AdDropdown)
    else:
        app.add_directive("ad-dropdown", EmptyDirective)
        app.add_directive("tab-set", EmptyDirective)
    app.connect("html-page-context", add_canonical_link)
    if hasattr(app, "add_css_file"):
        app.add_css_file("styles.css")
    else:
        global has_add_css_file
        has_add_css_file = False
        
    app.connect("builder-inited", _maybe_update_inventories)
    app.add_config_value(_UPDATE_KEY, default=False, rebuild=True, types=[bool])



from shrub.v3.evg_command import EvgCommandType
from shrub.v3.evg_command import expansions_update

from config_generator.etc.function import Function
from config_generator.etc.utils import bash_exec


class BootstrapMongoOrchestration(Function):
    name = 'bootstrap-mongo-orchestration'
    command_type = EvgCommandType.SETUP
    commands = [
        bash_exec(
            command_type=command_type,
            working_dir='mongoc',
            script='.evergreen/scripts/integration-tests.sh',
            add_expansions_to_env=True,
        ),
        expansions_update(
            command_type=command_type,
            file='mongoc/mo-expansion.yml'
        ),
    ]

    @classmethod
    def call(cls, **kwargs):
        return cls.default_call(**kwargs)


def functions():
    return BootstrapMongoOrchestration.defn()
